/*  miscimfns.h

    Mark Jenkinson, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

// Miscellaneous image functions


#if !defined(__miscimfns_h)
#define __miscimfns_h

#include <string>
#include "mjimage.h"
#include "miscmaths.h"
#include "lazy.h"

#ifndef NO_NAMESPACE
 using namespace NEWMAT;
 using namespace MISCMATHS;
 using namespace MJIMAGE;
 using namespace LAZY;

 namespace MISCIMFNS {
#endif

  ///////////////////////////////////////////////////////////////////////////

  // DIAGNOSTICS

  void print_dims(const volume& source);

  void print_all(const volume& source);

  void print_zeros(const volume& source);

  void print_volume_info(const volume& source, const string& name="");

  ///////////////////////////////////////////////////////////////////////////

  int get_min_max(const volume& v, float& max, float& min);

  int get_stats(const volume& v, float& sum, float& sum2);

  ColumnVector get_percentiles(const volume& v, const ColumnVector& percents);
  float get_percentile(const volume& v, const float p);

  inline bool quick_in_bounds(int x, int y, int z, int xb, int yb, int zb)
    {
      return ( (x>=0) && (y>=0) && (z>=0) && 
	       (x<xb) && (y<yb) && (z<zb) );
    }

  inline bool quick_max_bounds(int x, int y, int z, int xb, int yb, int zb)
    {
      return ((x<xb) && (y<yb) && (z<zb) );
    }

  inline bool equal_size(const volume& vol1, const volume& vol2)
    { return ( (vol2.slices()==vol1.slices()) && (vol2.rows()==vol1.rows()) &&
	       (vol2.columns()==vol1.columns()) ); }


  ///////////////////////////////////////////////////////////////////////////

  int corner_pad(const volume& vol, volume& target, float padval = 0.0);
  int corner_pad(volume& vol,int nx, int ny, int nz, float padval = 0.0);
  int centre_pad(const volume& vol, volume& target, int padding, 
		 float padval = 0.0);
  int centre_pad(volume& vol, int padding, float padval = 0.0);
  int pad(const volume& vol, volume& target, int padding, float padval);
  int pad(volume& vol, int nx, int ny, int nz, int padding, float padval);
  int unpad(const volume& vol, volume& target, int padding);
  int unpad(volume& vol, int padding);

  int clamp(volume& vol, float minval, float maxval);

  int blur4subsampling(volume& target, const volume& source,
		       const ColumnVector& final_vox_dims);
  int blur4subsampling(volume& target, const volume& source,
		       float final_iso_vox_dims);

  volume gaussian_kernel(float sigma, int radius);
  int make_grad_masks(volume& maskx, volume& masky, volume& maskz);
  int gradient(const volume& source, volume& grad, bool verbose=false);
  int log_edge_detect(volume& source, float sigma1, float sigma2);
  int fixed_edge_detect(volume& source, float toler);
  int fixed_edge_detect(volume& source);
  
  int convolve(const volume& source, volume& result, const volume& mask,
	       bool verbose=false);
  int mirror_convolve(const volume& source, volume& result, const volume& mask,
	       bool verbose=false);

  int isotropic_rescale(volume& iso, const volume& aniso,
			float scale, bool verbose=false);


  int subsample_by2(const volume& refvol, volume& halfvol);

  //----------------------------------------------------------------------//

  int affine_transform(volume& vout, const volume& vin, const Matrix& aff);


  int filled_affine_transform(volume& vout, const volume& vin, 
			      const Matrix& aff);

  class imagepair;
  int affine_transform(imagepair* ims, const Matrix& aff);



  ///////////////////////////////////////////////////////////////////////////

  class imagepair {
  public: 
    const volume &refvol;
    const volume &testvol;
    volume outputvol;
    int *bindex;
    float refmin;
    float refmax;
    float testmin;
    float testmax;
    float smoothsize;
    float fuzzyfrac;
    int count;
    int no_bins;
    ColumnVector plnp;
    int *jointhist;
    int *marghist1;
    int *marghist2;
    float *fjointhist;
    float *fmarghist1;
    float *fmarghist2;

    imagepair(volume& refv, volume& inputv) :
      refvol(refv), testvol(inputv), bindex(0), refmin(0.0), refmax(0.0), 
      testmin(0.0), testmax(0.0), smoothsize(1.0), fuzzyfrac(0.5), 
      count(0), no_bins(0)
      { 
       //if (!equal_size(refvol,testvol)) {
       //  cerr << "Unequally sized volumes to image pair: exiting" << endl;
       //  throw;
       //}
       //outputvol = refvol; // for cost evaluation want outputvol size = refvol
	
	get_min_max(refvol,refmin,refmax); 
	get_min_max(testvol,testmin,testmax); 
      }

    ~imagepair() { 
      if (jointhist)  delete [] jointhist; 
      if (marghist1)  delete [] marghist1; 
      if (marghist2)  delete [] marghist2; 
      if (fjointhist)  delete [] fjointhist; 
      if (fmarghist1)  delete [] fmarghist1; 
      if (fmarghist2)  delete [] fmarghist2; 
      if (bindex)     delete [] bindex;
    }
    

    void set_no_bins(int n_bins);

  };

  inline int *get_bindexptr(unsigned int x, unsigned int y, unsigned int z,
			    const volume& refvol, int *bindex) 
    {
      return bindex + ( z * refvol.ysize() + y ) * refvol.xsize() + x;
    }

  ///////////////////////////////////////////////////////////////////////////

  int calc_histogram(const volume& vol, int nbins, float minval, float maxval, 
		     ColumnVector& hist);
  
  
  int find_robust_limits(const volume& vol, int nbins, ColumnVector& hist,
			 float& minval, float& maxval);
  
  
  ///////////////////////////////////////////////////////////////////////////

#ifndef NO_NAMESPACE
 }
#endif

#endif





