/* {{{ Copyright etc. */

/*  avwmaths.c - basic image maths

    Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

/* }}} */
/* {{{ defines, includes and typedefs */

#include "libss/libss.h"
#include "libss/libavw.h"

/* }}} */
/* {{{ usage */

void usage(void)
{
  printf("\nUsage: avwmaths <first_input> [inputs and operations] <output>\n\n");
  printf("Note <first_input> becomes \"current image\" in descriptions below.\nSome inputs can be images or numbers.\n");
  printf("e.g. avwmaths grotin -add grotin2 grotout\nor   avwmaths grotin -add 2.5 grotout\n\n");
  printf("For two-image operations, the first may be t>1 and the second t=1.\n");
  printf("Operations:\n");
  printf("-add  : add following input to current image\n");
  printf("-sub  : subtract following input from current image\n");
  printf("-mul  : multiply current image by following input\n");
  printf("-div  : divide current image by following input\n");
  printf("-mas  : use (following image>0) to mask current image\n");
  printf("-thr  : use following number to threshold current image\n");
  printf("-bin  : use (current image>0) to binarise current image\n");
  printf("-dil  : use (current image>0) to dilate current image using 3x3x3 neighourhood\n");
  printf("-ero  : use (current image>0) to dilate current image using 3x3x3 neighourhood\n");
  exit(1);
}

/* }}} */
/* {{{ main */

int main(argc, argv)
  int   argc;
  char  *argv [];
{
  /* {{{ vars */

image_struct im, im2;
FDT  *in[1000];
int  i, j, x, y, z, t, xysize, xyzsize, size;
char *endptr;
double tmpd;

/* }}} */

  if (argc<3)
    usage();

  avw_read(argv[1],&im);
  in[0]=im.i;
  xysize=im.x*im.y;
  xyzsize=xysize*im.z;
  size=xyzsize*im.t;

  for (i = 2; i < argc-1; i++) {
    if (!strncmp(argv[i], "-mas", 4))
      /* {{{ mask */

{
  i++;
  avw_read(argv[i],&im2);
  in[1]=im2.i;
  for(j=0;j<size;j++)
    if (in[1][j%(im2.x*im2.y*im2.z*im2.t)]<=0)
      in[0][j]=0;
}

/* }}} */
    else if (!strncmp(argv[i], "-add", 4))
      /* {{{ add */

{
  i++;
  tmpd=strtod(argv[i],&endptr);
  if (endptr!=argv[i])
    for(j=0;j<size;j++)
      in[0][j]+=tmpd;
  else
    {
      avw_read(argv[i],&im2);
      in[1]=im2.i;
      for(j=0;j<size;j++)
	in[0][j]+=in[1][j%(im2.x*im2.y*im2.z*im2.t)];
    }
}

/* }}} */
    else if (!strncmp(argv[i], "-sub", 4))
      /* {{{ subtract */

{
  i++;
  tmpd=strtod(argv[i],&endptr);
  if (endptr!=argv[i])
    for(j=0;j<size;j++)
      in[0][j]-=tmpd;
  else
    {
      avw_read(argv[i],&im2);
      in[1]=im2.i;
      for(j=0;j<size;j++)
	in[0][j]-=in[1][j%(im2.x*im2.y*im2.z*im2.t)];
    }
}

/* }}} */
    else if (!strncmp(argv[i], "-mul", 4))
      /* {{{ multiply */

{
  i++;
  tmpd=strtod(argv[i],&endptr);
  if (endptr!=argv[i])
    for(j=0;j<size;j++)
      in[0][j]*=tmpd;
  else
    {
      avw_read(argv[i],&im2);
      in[1]=im2.i;
      for(j=0;j<size;j++)
	in[0][j]*=in[1][j%(im2.x*im2.y*im2.z*im2.t)];
    }
}

/* }}} */
    else if (!strncmp(argv[i], "-div", 4))
      /* {{{ divide */

{
  i++;
  tmpd=strtod(argv[i],&endptr);
  if (endptr!=argv[i])
    {
      if (tmpd!=0)
	for(j=0;j<size;j++)
	  in[0][j]/=tmpd;
      else
	for(j=0;j<size;j++)
	  in[0][j]=0;
    }
  else
    {
      avw_read(argv[i],&im2);
      in[1]=im2.i;
      for(j=0;j<size;j++)
	if (in[1][j]!=0)
	  in[0][j]/=in[1][j];
	else
	  in[0][j]=0;
    }
}

/* }}} */
    else if (!strncmp(argv[i], "-thr", 4))
      /* {{{ threshold */

{
  i++;
  tmpd=strtod(argv[i],&endptr);
  for(j=0;j<size;j++)
    if (in[0][j]<tmpd)
      in[0][j]=0;
}

/* }}} */
    else if (!strncmp(argv[i], "-bin", 4))
      /* {{{ binarise */

{
  for(j=0;j<size;j++)
    if (in[0][j]>0)
      in[0][j]=1;
    else
      in[0][j]=0;
}

/* }}} */
    else if (!strncmp(argv[i], "-dil", 4))
      /* {{{ dilate */

{
  FDT *tmpim=malloc(size*sizeof(FDT));
  memcpy(tmpim,in[0],size*sizeof(FDT));

  for(t=0;t<im.t;t++) for(z=0;z<im.z;z++) for(y=0;y<im.y;y++) for(x=0;x<im.x;x++)
    if (in[0][t*xyzsize+z*xysize+y*im.x+x]==0)
      {
	int xx, yy, zz, count=0;
	double total=0;
	for(zz=MAX(z-1,0);zz<=MIN(z+1,im.z-1);zz++)
	  for(yy=MAX(y-1,0);yy<=MIN(y+1,im.y-1);yy++)
	    for(xx=MAX(x-1,0);xx<=MIN(x+1,im.x-1);xx++)
	      if (in[0][t*xyzsize+zz*xysize+yy*im.x+xx]!=0)
		{
		  total+=in[0][t*xyzsize+zz*xysize+yy*im.x+xx];
		  count++;
		}
	if (count>0)
	  tmpim[t*xyzsize+z*xysize+y*im.x+x]=total/count;
      }

  free(in[0]);
  in[0]=tmpim;
}

/* }}} */
    else if (!strncmp(argv[i], "-ero", 4))
      /* {{{ erode */

{
  FDT *tmpim=malloc(size*sizeof(FDT));
  memcpy(tmpim,in[0],size*sizeof(FDT));

  for(t=0;t<im.t;t++) for(z=0;z<im.z;z++) for(y=0;y<im.y;y++) for(x=0;x<im.x;x++)
    if (in[0][t*xyzsize+z*xysize+y*im.x+x]!=0)
      {
	int xx, yy, zz;
	for(zz=MAX(z-1,0);zz<=MIN(z+1,im.z-1);zz++)
	  for(yy=MAX(y-1,0);yy<=MIN(y+1,im.y-1);yy++)
	    for(xx=MAX(x-1,0);xx<=MIN(x+1,im.x-1);xx++)
	      if (in[0][t*xyzsize+zz*xysize+yy*im.x+xx]==0)
		tmpim[t*xyzsize+z*xysize+y*im.x+x]=0;
      }

  free(in[0]);
  in[0]=tmpim;
}

/* }}} */
    else
      usage();
  }

  im.i=in[0];
  find_thresholds(&im,0.1);
  avw_write(argv[i],im);
  return(0);
}

/* }}} */
