/* {{{ Copyright etc. */

/*  tsplot - FMRI time series and model plotting

    Stephen Smith and Mark Woolrich, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

/* }}} */
/* {{{ defines, includes and typedefs */

#include "newimageall.h"
#include "newmatap.h"
#include "miscmaths.h"
#include <vector>

using namespace NEWMAT;
using namespace NEWIMAGE;

/* }}} */
/* {{{ usage */

void usage(void)
{
  printf("Usage: tsplot <feat_directory.feat> [options]\n");
  printf("[-f <4D_data>] input main filtered data, in case it's not in the FEAT directory\n");
  printf("[-p <+_string>] string of pluses to insert\n");
  printf("[-c <X Y Z>] : use X,Y,Z instead of max Z stat position\n");
  printf("[-C <X Y Z output_file.txt>] : use X,Y,Z to output time series only - no stats or modelling\n");
  printf("[-o <output_directory>] change output directory from default of input feat directory\n");
  exit(1);
}

/* }}} */
/* {{{ find_key */

char *find_key(FILE *fd, char *control_line, const char *key)
{
  while (fgets(control_line, 1000, fd)!=NULL)
    if ( (strncmp(control_line,key,strlen(key))==0) )
      {
	return control_line+strlen(key);
      }
	  
  printf("Error: key \"%s\" not found.\n",key);
  exit(1);
}

/* }}} */
/* {{{ establish_pwfilter */

void establish_pwfilter(const ColumnVector& ac, ColumnVector& pwfilter, int zeropad, int npts)
{
  // FFT auto corr estimate
  ColumnVector dummy(zeropad);
  dummy = 0;
  ColumnVector vrow(zeropad);
  vrow = 0;

  // ac maybe cutoff to be smaller than npts
  int sizeTS = ac.Nrows();
  if (sizeTS > npts/2)
    sizeTS = npts/2;
  
  vrow.Rows(1,sizeTS) = ac.Rows(1,sizeTS);
  vrow.Rows(zeropad - sizeTS + 2, zeropad) = ac.Rows(2, sizeTS).Reverse();

  ColumnVector ac_fft_imag;
  FFT(vrow, dummy, pwfilter, ac_fft_imag);      

  // inverse auto corr to give prewhitening filter
  // no DC component so set first value to 0
  pwfilter(1) = 0.0;
  
  for(int j = 2; j <= zeropad; j++)
      pwfilter(j) = 1.0/sqrt(pwfilter(j));	      
  
  // normalise pwfilter such that sum(j)((pwfilter)^2/zeropad)) = 1
  pwfilter /= sqrt(pwfilter.SumSquare()/zeropad);
}

/* }}} */
/* {{{ prewhiten_data */

void prewhiten_data(const ColumnVector& data, ColumnVector& pwdata, ColumnVector& pwfilter, int zeropad, int npts)
{ 
  ColumnVector data_fft_real, data_fft_imag, realifft, dummy;
  dummy.ReSize(zeropad);
  dummy = 0;

  // Remove and store mean
  float mn = MISCMATHS::mean(data);
  pwdata.ReSize(zeropad);
  pwdata = 0;
  pwdata.Rows(1,npts) = data - mn;

  // FFT data
  FFT(pwdata, dummy, data_fft_real, data_fft_imag);
  FFTI(SP(pwfilter, data_fft_real), SP(pwfilter, data_fft_imag), realifft, dummy); 
  // take first npts and restore mean
  pwdata = realifft.Rows(1,npts) + mn;  
}

/* }}} */
/* {{{ prewhiten_model */

void prewhiten_model(const ColumnVector& ac, double *model, double *pwmodel, int nevs, int npts)
{
  Matrix dm(npts, nevs), pwdm(npts, nevs);

  // Get values out of model
  for(int ev=0; ev<nevs; ev++)
    {
      for(int t=0; t<npts; t++)
	{
	  dm(t+1,ev+1) = model[t*nevs+ev];
	}
    }

  pwdm = dm;

  int zeropad = (int)pow(2,ceil(log(npts)/log(2)));
  
  // get prewhitening filter from ac
  ColumnVector pwfilter;
  establish_pwfilter(ac, pwfilter, zeropad, npts);

  // prewhiten each of the evs
  for(int ev = 1; ev <= nevs; ev++)
    {
      ColumnVector pwdata;
      prewhiten_data(dm.Column(ev), pwdata, pwfilter, zeropad, npts);
      pwdm.Column(ev) = pwdata;
    }

  // setup and return result
  for(int ev=0; ev<nevs; ev++)
    {
      for(int t=0; t<npts; t++)
	{
	  pwmodel[t*nevs+ev] = pwdm(t+1,ev+1);
	}
    }
}

/* }}} */
/* {{{ prewhiten_timeseries */

void prewhiten_timeseries(const ColumnVector& ac, const ColumnVector& ts, ColumnVector& pwts, int npts)
{
  int zeropad = (int)pow(2,ceil(log(npts)/log(2)));
  
  // get prewhitening filter from ac
  ColumnVector pwfilter;
  establish_pwfilter(ac, pwfilter, zeropad, npts);

  // prewhiten
  prewhiten_data(ts, pwts, pwfilter, zeropad, npts);
}

/* }}} */
/* {{{ read_model */

double *read_model(char *filename, int *nevs, int *npts)
{
  FILE *designfp;
  char control_line[1010];
  double *model;
  int i;

  if((designfp=fopen(filename,"r"))==NULL)
    {
      *npts=0;
      return NULL;
    }

  *nevs=(int)atof(find_key(designfp,control_line,"/NumWaves"));
  *npts=(int)atof(find_key(designfp,control_line,"/NumPoints"));

  atof(find_key(designfp,control_line,"/Matrix"));

  model=(double*)malloc(sizeof(double)* *nevs * *npts);

  for(i=0; i<*nevs * *npts; i++)
    fscanf(designfp,"%lf",&model[i]);

  return model;
}

/* }}} */
/* {{{ read_contrasts */

double *read_contrasts(char *filename, int *nevs, int *ncon)
{
  FILE *designfp;
  char control_line[1010];
  double *contrasts;
  int i;

  if((designfp=fopen(filename,"r"))==NULL)
    {
      fprintf(stderr,"Can't open %s\n",filename);
      exit(1);
    }

  *nevs=(int)atof(find_key(designfp,control_line,"/NumWaves"));
  *ncon=(int)atof(find_key(designfp,control_line,"/NumContrasts"));

  atof(find_key(designfp,control_line,"/Matrix"));

  contrasts=(double*)malloc(sizeof(double)* *nevs * *ncon);

  for(i=0; i<*nevs * *ncon; i++)
    fscanf(designfp,"%lf",&contrasts[i]);

  return contrasts;
}

/* }}} */
/* {{{ main */

int main(int argc, char **argv)
{
  /* {{{ variables */

FILE         *ifp, *ofp;
int          argi=1, prewhiten=1, ev, i, t, X=0, Y=0, Z=0, nevs, npts, ncon=1, size, ymin, ymax, coordset=0, dataonly=0, modelfree=0;
double       *model, *contrasts=NULL, tsmean, tmpf, maxz;
char         fmridata[10000], featdir[10000], outputdir[10000], plusstring[10000], gpstart[10000],
             thestring[10000], thestring2[10000], thestring3[10000], datafile[10000], binstr[10000];
ColumnVector pwts;

/* }}} */

  /* {{{ process arguments */

if (argc<2) usage();

strcpy(featdir,argv[argi++]);

strcpy(outputdir,featdir);

sprintf(fmridata,"%s/filtered_func_data",featdir);

/* will use binstr to work out where "gnuplot" and "convert" are, for later system calls */
sprintf(binstr,"%s",argv[0]);
binstr[strlen(binstr)-strlen("tsplot")]=0;

for (;argi<argc;argi++)
{
  if (!strcmp(argv[argi], "-f"))
    /* {{{ alternative fmri data */

{
  argi++;
  if (argc<argi+1)
    {
      printf("Error: no value given following -f\n");
      usage();
    }
  strcpy(fmridata,argv[argi]);
}

/* }}} */
  if (!strcmp(argv[argi], "-p"))
    /* {{{ plus string */

{
  argi++;
  if (argc<argi+1)
    {
      printf("Error: no value given following -p\n");
      usage();
    }
  strcpy(plusstring,argv[argi]);
}

/* }}} */
  else if (!strcmp(argv[argi], "-c"))
    /* {{{ alternative voxel position */

{
  coordset=1;
  argi++;
  if (argc<argi+3) /* options following c haven't been given */
    {
      printf("Error: incomplete values given following -c\n");
      usage();
    }
  X=atoi(argv[argi++]);
  Y=atoi(argv[argi++]);
  Z=atoi(argv[argi]);
}

/* }}} */
  else if (!strcmp(argv[argi], "-C"))
    /* {{{ output data only */

{
  coordset=1;
  dataonly=1;
  argi++;
  if (argc<argi+4)
    {
      printf("Error: incomplete values given following -C\n");
      usage();
    }
  X=atoi(argv[argi++]);
  Y=atoi(argv[argi++]);
  Z=atoi(argv[argi++]);
  strcpy(datafile,argv[argi]);
}

/* }}} */
  if (!strcmp(argv[argi], "-o"))
    /* {{{ output dir */

{
  argi++;
  if (argc<argi+1)
    {
      printf("Error: no value given following -o\n");
      usage();
    }
  strcpy(outputdir,argv[argi]);
}

/* }}} */
}

/* }}} */
  /* {{{ read filtered_func_data */

volume4D<int> im;
read_volume4D(im, fmridata);

/*printf("FMRI dimensions %d %d %d %d\n",image.x,image.y,image.z,image.t);*/

size=im.nvoxels();

if (dataonly)
  /* {{{ output raw data and exit */

{
  if((ofp=fopen(datafile,"w"))==NULL)
    {
      fprintf(stderr,"Can't open output data file %s\n",datafile);
      exit(1);
    }

  for(t=0; t<im.tsize(); t++)
    fprintf(ofp,"%e\n",(float)im(X,Y,Z,t));
  //[t*size+Z*image.x*image.y+Y*image.x+X]);

  fclose(ofp);

  return 0;
}

/* }}} */

/* }}} */
  /* {{{ read design.mat */

sprintf(thestring,"%s/design%s.mat",featdir,plusstring);
model=read_model(thestring,&nevs,&npts);

if (npts==0) modelfree=1;

npts=im.tsize();

/* }}} */
  /* {{{ read auto correlation estimates for prewhitening */

double* pwmodel=0;
volume4D<float> acs;
sprintf(thestring,"%s/stats/threshac1.hdr",featdir);
if((ifp=fopen(thestring,"r"))!=NULL)
{
  fclose(ifp);
  sprintf(thestring,"%s/stats/threshac1",featdir);
  read_volume4D(acs, thestring);
  // allocate prewhiten variable for later use
  pwmodel=new double[nevs*npts];
}
else
{ /* no autocorr estimates - thus will use the uncorrected design matrix */
  prewhiten=0;
}

/* }}} */
  /* {{{ read design.con and PEs */

vector< volume<float> > impe(nevs);
if (!modelfree)
{
  sprintf(thestring,"%s/design%s.con",featdir,plusstring);
  contrasts=read_contrasts(thestring,&nevs,&ncon);

  for(i=1;i<=nevs;i++)
    {
      sprintf(thestring,"%s/stats/pe%d",featdir,i);
      /*printf("Reading PE image %s\n",thestring);*/
      read_volume(impe[i-1], thestring);
    }
}

/* }}} */
  /* {{{ create plot(s) for each contrast */

for(i=0; i<ncon; i++)
{
  double *data=(double*)calloc(npts,sizeof(double));

  volume<float> imcope;
  volume<float> imz;
  volume<float> imclus;

  bool haveclusters=false;

  /* {{{ read COPE and derived stats */
{
  FILE *ifp;

  if (!modelfree)
    {
      sprintf(thestring,"%s/stats/cope%s%d",featdir,plusstring,i+1);
      /*printf("Reading COPE image %s\n",thestring);*/
      read_volume(imcope,thestring);
    }

  sprintf(thestring,"%s/stats/zstat%s%d",featdir,plusstring,i+1);
  /*printf("Reading Z stat image %s\n",thestring);*/
  read_volume(imz,thestring);

  if (!coordset)
    {
      sprintf(thestring,"%s/cluster_mask_zstat%s%d.hdr",featdir,plusstring,i+1);
      if((ifp=fopen(thestring,"r"))!=NULL)
	{
	  fclose(ifp);
	  sprintf(thestring,"%s/cluster_mask_zstat%s%d",featdir,plusstring,i+1);
	  /*printf("Reading cluster image %s\n",thestring);*/
	  read_volume(imclus,thestring);
	  haveclusters=(imclus.max()>0);
	}
    }
}

/* }}} */

  /* {{{ open data file */

sprintf(thestring,"%s/tsplot%s%d.txt",outputdir,plusstring,i+1);

if((ofp=fopen(thestring,"w"))==NULL)
{
  fprintf(stderr,"Can't open output data file %s\n",thestring);
  exit(1);
}

/* }}} */
  /* {{{ find max Z and X,Y,Z */

if (!coordset)
{
  X=Y=Z=0;
  maxz=0;

  for(int z=0; z<im.zsize(); z++)
    for(int y=0; y<im.ysize(); y++)
      for(int x=0; x<im.xsize(); x++)
	if ( (imz(x,y,z)>maxz) &&
	     ( (!haveclusters) ||   /* make max Z be inside a cluster if we found a cluster map */
	       (imclus(x,y,z)>0) && (!prewhiten || acs(x,y,z,1)!=0) ) )
	  {
	    maxz=imz(x,y,z);
	    X=x; Y=y; Z=z;
	  }

  /*printf("Z stat %d has maximum of %f at %d %d %d\n",i+1,maxz,X,Y,Z);*/
}
else
  maxz=imz(X,Y,Z);

/* }}} */
  /* {{{ create 1D data vector and find mean */

tsmean=0;

if(prewhiten)
     prewhiten_timeseries(acs.voxelts(X,Y,Z), im.voxelts(X,Y,Z), pwts, npts);
else
     pwts = im.voxelts(X,Y,Z);

for(t=0; t<npts; t++)
{
  data[t]=pwts(t+1);
  tsmean+=data[t];
}

tsmean/=npts;

/* }}} */
  /* {{{ output the data and model */

ymin=ymax=(int)data[0];

if (prewhiten)
     prewhiten_model(acs.voxelts(X,Y,Z), model, pwmodel, nevs, npts);

for(t=0; t<npts; t++)
{
  double fullmodel=0, copemodel=0;

  if (!modelfree)
    for(ev=0; ev<nevs; ev++)
      {
	if (prewhiten)
	  tmpf=pwmodel[t*nevs+ev]*impe[ev](X,Y,Z);
	else
	  tmpf=model[t*nevs+ev]*impe[ev](X,Y,Z);
	fullmodel += tmpf;
	copemodel += tmpf*contrasts[i*nevs+ev];
      }
      
  fprintf(ofp,"%e ",data[t]);
  ymin=(int)MISCMATHS::Min(data[t],ymin);
  ymax=(int)MISCMATHS::Max(data[t],ymax);

  if (!modelfree)
    {
      fprintf(ofp,"%e ",tsmean+copemodel);
      ymin=(int)MISCMATHS::Min(tsmean+copemodel,ymin);
      ymax=(int)MISCMATHS::Max(tsmean+copemodel,ymax);

      fprintf(ofp,"%e ",tsmean+fullmodel);
      ymin=(int)MISCMATHS::Min(tsmean+fullmodel,ymin);
      ymax=(int)MISCMATHS::Max(tsmean+fullmodel,ymax);

      fprintf(ofp,"%e ",data[t]-fullmodel+copemodel);
    }

  fprintf(ofp,"\n");
}

ymax+=(ymax-ymin)/5;
ymin-=(ymax-ymin)/20;

fclose(ofp);

/* }}} */
  if (!modelfree)
    /* {{{ call gnuplot via system call */

{

thestring2[0]=0;
for(ev=0;ev<nevs;ev++)
{
  if ((float)((int)contrasts[i*nevs+ev])==contrasts[i*nevs+ev])
    sprintf(thestring2,"%s %d",thestring2,(int)contrasts[i*nevs+ev]);
  else
    sprintf(thestring2,"%s %.1f",thestring2,contrasts[i*nevs+ev]);
}

sprintf(gpstart,"echo \"set term pbm color\nset size %f,0.5\nset title 'Contrast %d [%s ] :",MISCMATHS::Max(npts*4,600)/640.0,i+1,thestring2);

sprintf(thestring2,"%s/tsplot%s%d",outputdir,plusstring,i+1);

if (!coordset)
     sprintf(thestring3,"%s max Z stat of %.3f at voxel",gpstart,maxz);
else
     sprintf(thestring3,"%s Z stat of %.3f at selected voxel",gpstart,maxz);

sprintf(thestring,"%s (%d,%d,%d)'\nset yrange [%d:%d]\nplot '%s.txt' using 1 title 'data' with lines , '%s.txt' using 2 title 'cope partial model fit' with lines , '%s.txt' using 3 title 'full model fit' with lines \" | %sgnuplot | %sconvert - %s.gif",thestring3,X,Y,Z,ymin,ymax,thestring2,thestring2,thestring2,binstr,binstr,thestring2);

system(thestring);


sprintf(thestring,"%s (%d,%d,%d)'\nset yrange [%d:%d]\nplot '%s.txt' using 4 title 'reduced data' with lines , '%s.txt' using 2 title 'cope partial model fit' with lines \" | %sgnuplot | %sconvert - %sp.gif",thestring3,X,Y,Z,ymin,ymax,thestring2,thestring2,binstr,binstr,thestring2);

system(thestring);

}

/* }}} */
  else
    /* {{{ call gnuplot via system call */

{
  sprintf(gpstart,"echo \"set term pbm color\nset size %f,0.5\nset title '",MISCMATHS::Max(npts*4,600)/640.0);

  sprintf(thestring2,"%s/tsplot%s%d",outputdir,plusstring,i+1);
  
  if (!coordset)
    sprintf(thestring3,"%s max Z stat of %.3f at voxel",gpstart,maxz);
  else
    sprintf(thestring3,"%s Z stat of %.3f at selected voxel",gpstart,maxz);

  sprintf(thestring,"%s (%d,%d,%d)'\nset yrange [%d:%d]\nplot '%s.txt' using 1 title 'data' with lines \" | %sgnuplot | %sconvert - %s.gif",thestring3,X,Y,Z,ymin,ymax,thestring2,binstr,binstr,thestring2);
 
  system(thestring);
}

/* }}} */

  if (haveclusters)
    {
      double *fullmodel=(double*)calloc(npts,sizeof(double)),
	*copemodel=(double*)calloc(npts,sizeof(double)),
	*data=(double*)calloc(npts,sizeof(double)),
	ztotal=0;
      int count=0;

      /* {{{ open data file */

sprintf(thestring,"%s/tsplotc%s%d.txt",outputdir,plusstring,i+1);

if((ofp=fopen(thestring,"w"))==NULL)
{
  fprintf(stderr,"Can't open output data file %s\n",thestring);
  exit(1);
}

/* }}} */
      /* {{{ COMMENT UNWEIGHTED create 1D data vector and find mean */

#ifdef FoldingComment

 for(int x=0; x<im.xsize(); x++)
   for(int y=0; y<im.ysize(); y++)
     for(int z=0; z<im.zsize(); z++)
       if (imclus(x,y,z)>0 && (!prewhiten || acs(x,y,z,1)!=0))
	 {
	   count++;
	   if(prewhiten)
	     prewhiten_timeseries(acs.voxelts(z,y,z), im.voxelts(z,y,z), pwts, npts);
	   else
	     pwts = im.voxelts(x,y,z);

	   for(t=0; t<npts; t++)
	     data[t]+=pwts(t+1);
	 }
 
/*printf("count=%d\n",count);*/

tsmean=0;
for(t=0; t<npts; t++)
{
  data[t]/=count;
  tsmean+=data[t];
}
tsmean/=npts;

ymin=ymax=(int)data[0];

if (!modelfree)
  for(int x=0; x<im.xsize(); x++)
    for(int y=0; y<im.ysize(); y++)
      for(int z=0; z<im.zsize(); z++)
     if (imclus(x,y,z)>0 && (!prewhiten || acs(x,y,z,1) != 0))
	  {
	    if (prewhiten)
	      pwmodel = prewhiten_model(acs.voxelts(x,y,z), model, pwmodel, nevs, npts);

	    for(t=0; t<npts; t++)
	      for(ev=0; ev<nevs; ev++)
		{
		  if (prewhiten)
		    tmpf=pwmodel[t*nevs+ev]*impe[ev](x,y,z);
		  else
		    tmpf=model[t*nevs+ev]*impe[ev](x,y,z);
		  fullmodel[t] += tmpf;
		  copemodel[t] += tmpf*contrasts[i*nevs+ev];
		}
	    //	    cerr << "acs.voxelts(x,y,z)(1)=" << acs.voxelts(x,y,z)(1) << "imclus(x,y,z)=" << imclus(x,y,z) << ",pwmodel[80]=" << pwmodel[80]<< ",*impe[1](x,y,z)=" << impe[1](x,y,z) << endl;
	  }

for(t=0; t<npts; t++)
{
  if (!modelfree)
    {
      fullmodel[t] /= count;
      copemodel[t] /= count;
    }
  
  fprintf(ofp,"%e ",data[t]);
  ymin=(int)MISCMATHS::Min(data[t],ymin);
  ymax=(int)MISCMATHS::Max(data[t],ymax);

  if (!modelfree)
    {
      fprintf(ofp,"%e ",tsmean+copemodel[t]);
      ymin=(int)MISCMATHS::Min(tsmean+copemodel[t],ymin);
      ymax=(int)MISCMATHS::Max(tsmean+copemodel[t],ymax);

      fprintf(ofp,"%e ",tsmean+fullmodel[t]);
      ymin=(int)MISCMATHS::Min(tsmean+fullmodel[t],ymin);
      ymax=(int)MISCMATHS::Max(tsmean+fullmodel[t],ymax);
	
      fprintf(ofp,"%e ",data[t]-fullmodel[t]+copemodel[t]);
    }

  fprintf(ofp,"\n");
}

ymax+=(ymax-ymin)/5;
ymin-=(ymax-ymin)/20;

fclose(ofp);

#endif

/* }}} */
      /* {{{ create 1D data vector and find mean */

 for(int x=0; x<im.xsize(); x++)
   for(int y=0; y<im.ysize(); y++)
     for(int z=0; z<im.zsize(); z++)
       if (imclus(x,y,z)>0 && (!prewhiten || acs(x,y,z,1)!=0))
	 {
	   count++;
	   ztotal+=imz(x,y,z);
	   if(prewhiten)
	     prewhiten_timeseries(acs.voxelts(x,y,z), im.voxelts(x,y,z), pwts, npts);
	   else
	     pwts = im.voxelts(x,y,z);

	   for(t=0; t<npts; t++)
	     data[t]+=pwts(t+1)*imz(x,y,z);
	 }
 
/*printf("count=%d\n",count);*/

tsmean=0;
for(t=0; t<npts; t++)
{
  data[t]/=ztotal;
  tsmean+=data[t];
}
tsmean/=npts;

ymin=ymax=(int)data[0];

if (!modelfree)
  for(int x=0; x<im.xsize(); x++)
    for(int y=0; y<im.ysize(); y++)
      for(int z=0; z<im.zsize(); z++)
     if (imclus(x,y,z)>0 && (!prewhiten || acs(x,y,z,1) != 0))
	  {
	    if (prewhiten)
	      prewhiten_model(acs.voxelts(x,y,z), model, pwmodel, nevs, npts);

	    for(t=0; t<npts; t++)
	      for(ev=0; ev<nevs; ev++)
		{
		  if (prewhiten)
		    tmpf=pwmodel[t*nevs+ev]*impe[ev](x,y,z);
		  else
		    tmpf=model[t*nevs+ev]*impe[ev](x,y,z);
		  fullmodel[t] += tmpf*imz(x,y,z);
		  copemodel[t] += tmpf*contrasts[i*nevs+ev]*imz(x,y,z);
		}
	  }

for(t=0; t<npts; t++)
{
  if (!modelfree)
    {
      fullmodel[t] /= ztotal;
      copemodel[t] /= ztotal;
    }
  
  fprintf(ofp,"%e ",data[t]);
  ymin=(int)MISCMATHS::Min(data[t],ymin);
  ymax=(int)MISCMATHS::Max(data[t],ymax);

  if (!modelfree)
    {
      fprintf(ofp,"%e ",tsmean+copemodel[t]);
      ymin=(int)MISCMATHS::Min(tsmean+copemodel[t],ymin);
      ymax=(int)MISCMATHS::Max(tsmean+copemodel[t],ymax);

      fprintf(ofp,"%e ",tsmean+fullmodel[t]);
      ymin=(int)MISCMATHS::Min(tsmean+fullmodel[t],ymin);
      ymax=(int)MISCMATHS::Max(tsmean+fullmodel[t],ymax);
	
      fprintf(ofp,"%e ",data[t]-fullmodel[t]+copemodel[t]);
    }

  fprintf(ofp,"\n");
}

ymax+=(ymax-ymin)/5;
ymin-=(ymax-ymin)/20;

fclose(ofp);

/* }}} */

      if (!modelfree)
	/* {{{ call gnuplot via system call */
{
sprintf(thestring2,"%s/tsplotc%s%d",outputdir,plusstring,i+1);

sprintf(thestring,"%s averaged over %d voxels'\nset yrange [%d:%d]\nplot '%s.txt' using 1 title 'data' with lines , '%s.txt' using 2 title 'cope partial model fit' with lines , '%s.txt' using 3 title 'full model fit' with lines \" | %sgnuplot | %sconvert - %s.gif",gpstart,count,ymin,ymax,thestring2,thestring2,thestring2,binstr,binstr,thestring2);

system(thestring);

sprintf(thestring,"%s averaged over %d voxels'\nset yrange [%d:%d]\nplot '%s.txt' using 4 title 'reduced data' with lines , '%s.txt' using 2 title 'cope partial model fit' with lines \" | %sgnuplot | %sconvert - %sp.gif",gpstart,count,ymin,ymax,thestring2,thestring2,binstr,binstr,thestring2);

system(thestring);
}

/* }}} */
      else
	/* {{{ call gnuplot via system call */

{
  sprintf(thestring2,"%s/tsplotc%s%d",outputdir,plusstring,i+1);

  sprintf(thestring,"%s averaged over %d voxels'\nset yrange [%d:%d]\nplot '%s.txt' using 1 title 'data' with lines \" | %sgnuplot | %sconvert - %s.gif",gpstart,count,ymin,ymax,thestring2,binstr,binstr,thestring2);

  system(thestring);
}

/* }}} */
    }
}

/* }}} */
  /* {{{ clean up and exit */

  delete [] pwmodel;
  return 0;

/* }}} */
}

/* }}} */
