/*  miscimfns.cc

    Mark Jenkinson, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

// Miscellaneous image functions


#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm>
#include "interpolation.h"
#include "miscimfns.h"
#include "math.h"

#ifndef NO_NAMESPACE
 using namespace MISCMATHS;
 using namespace INTERPOLATION;
 using namespace MJIMAGE;

 namespace MISCIMFNS {
#endif

  // DIAGNOSTICS

  void print_dims(const volume& source)
    {
      cout << "Dims: " << source.columns() << ", " << source.rows()
	   << ", " << source.slices() << endl;
    }


  void print_all(const volume& source)
    {
      unsigned int zb = source.slices(), yb = source.rows(), 
	xb = source.columns();
      for (unsigned int z=0; z<zb; z++) { 
	for (unsigned int y=0; y<yb; y++) { 
	  for (unsigned int x=0; x<xb; x++) { 
	    cout << source(x,y,z) << " ";
	  }
	}
      }
    }

  void print_zeros(const volume& source)
    {
      unsigned int zb = source.slices(), yb = source.rows(), 
	xb = source.columns();
      for (unsigned int z=0; z<zb; z++) { 
	for (unsigned int y=0; y<yb; y++) { 
	  for (unsigned int x=0; x<xb; x++) { 
	    if (source(x,y,z) == 0.0)  
	      cout << x << "," << y << "," << z << " -> " 
		   << source(x,y,z) << endl;
	  }
	}
      }
    }

  void print_volume_info(const volume& source, const string& name)
    {
      float fmin, fmax;
      cout << name << ":: Size = (" << source.columns() << ","
	   << source.rows() << "," << source.slices() << ")" << endl;
      cout << name << ":: Dims = (" << source.getx() << ","
	   << source.gety() << "," << source.getz() << ")" << endl;
      get_min_max(source,fmin,fmax);
      cout << name << ":: Maximum and minimum intensities are: " 
	   << fmax << " and " << fmin << endl;
    }  

  
  ////////////////////////////////////////////////////////////////////////

  int get_min_max(const volume& v, float& min, float& max)
    {
      max=v(0,0,0);
      min=v(0,0,0);
      float val;
      unsigned int zb = v.slices(), yb = v.rows(), xb = v.columns();
      for (unsigned int z=0; z<zb; z++) { 
	for (unsigned int y=0; y<yb; y++) { 
	  for (unsigned int x=0; x<xb; x++) { 
	    val = v(x,y,z);
	    if (val < min)  min=val;
	    if (val > max)  max=val;
	  }
	}
      }
      return 0;
    }


  int get_stats(const volume& v, float& sum, float& sum2)
    {
      sum=0.0;
      sum2=0.0;
      float val;
      unsigned int zb = v.slices(), yb = v.rows(), xb = v.columns();
      for (unsigned int z=0; z<zb; z++) { 
	for (unsigned int y=0; y<yb; y++) { 
	  for (unsigned int x=0; x<xb; x++) { 
	    val = v(x,y,z);
	    sum += val;
	    sum2 += val*val;
	  }
	}
      }
      return 0;
    }



  ColumnVector get_percentiles(const volume& v, const ColumnVector& percents)
    {
      // All percents should be in [0,100] NOT [0,1]
      unsigned int zb = v.slices(), yb = v.rows(), xb = v.columns();
      unsigned int numbins = xb*yb*zb;
      unsigned int hindx = 0;
      std::vector<float> hist(numbins);
      for (unsigned int z=0; z<zb; z++) {
	for (unsigned int y=0; y<yb; y++) {
	  for (unsigned int x=0; x<xb; x++) {
	    hist[hindx++] = v(x,y,z);
	  }
	}
      }
      sort(hist.begin(),hist.end());

      ColumnVector outputvals(percents.Nrows());
      for (int n=1; n<=percents.Nrows(); n++) {
	unsigned int percentile = 
	  (unsigned int) (((float) numbins) * percents(n) / 100.0);
	if (percentile<0)  percentile=0;
	if (percentile>=numbins)  percentile=numbins-1;
	outputvals(n) = hist[percentile];
      }
      return outputvals;
    }      

     
  float get_percentile(const volume& v, const float p)
    {
      ColumnVector percents(1), outputval;
      percents(1) = p;
      outputval = get_percentiles(v,percents);
      if (outputval.Nrows()<1) {
	cerr << "WARNING (get_percentile): did not calculate any percentiles" 
	     << endl;
      }
      return outputval(1);
    }


///////////////////////////////////////////////////////////////////////////

// More complicated, image functions

  int corner_pad(const volume& vol, volume& target, float padval)
    {
      // copies the volume vol to target, padding any excess with padval
      //  note that the copied volume starts at (0,0,0) in the target
      // Note: the size of the target volume must be set beforehand
      target = padval;
      unsigned int bx, by, bz;
      bx = Min(target.columns(),vol.columns());
      by = Min(target.rows(),vol.rows());
      bz = Min(target.slices(),vol.slices()); 
      for (unsigned int z=0; z<bz; z++) {
	for (unsigned int y=0; y<by; y++) {
	  for (unsigned int x=0; x<bx; x++) {
	    target(x,y,z) = vol(x,y,z);
	  }
	}
      }
      return 0;
    }

  int corner_pad(volume& vol, int nx, int ny, int nz, float padval)
    {
      // resizes this volume, and pads it with padval...
      if ( (vol.slices()==nz) && (vol.rows()==ny) && (vol.columns()==nx) )
	return 0;
      volume workingvol(nx,ny,nz);
      corner_pad(vol,workingvol,padval);
      vol.reinitialize(workingvol);
      return 0;
    }


  int centre_pad(const volume& vol, volume& target, int padding, float padval)
    {
      // copies the volume vol to target, padding any excess with padval
      //  note that the copied volume starts at (0,0,0) in the target
      unsigned int bx, by, bz;
      bx = vol.columns() + 2*padding;
      by = vol.rows() + 2*padding;
      bz = vol.slices() + 2*padding; 
      target.reinitialize(bx,by,bz);
      target = padval;
      for (unsigned int z=padding; z<bz-padding; z++) {
	for (unsigned int y=padding; y<by-padding; y++) {
	  for (unsigned int x=padding; x<bx-padding; x++) {
	    target(x,y,z) = vol(x-padding,y-padding,z-padding);
	  }
	}
      }
      ColumnVector voxorig;
      voxorig = target.getvoxelorigin();
      voxorig += (float) padding;
      target.setvoxelorigin(voxorig);
      return 0;
    }


  int centre_pad(volume& vol, int padding, float padval)
    {
      if (padding <= 0) return 0;
      volume workingvol;
      centre_pad(vol,workingvol,padding,padval);
      vol.reinitialize(workingvol);
      return 0;
    }


  int pad(const volume& vol, volume& target, int padding, float padval)
    {
      // copies the volume vol to target, padding any excess with padval
      //  note that the copied volume starts at (0,0,0) in the target
      unsigned int bx, by, bz;
      bx = Min(target.columns(),vol.columns()+padding);
      by = Min(target.rows(),vol.rows()+padding);
      bz = Min(target.slices(),vol.slices()+padding); 
      target = padval;
      for (unsigned int z=padding; z<bz; z++) {
	for (unsigned int y=padding; y<by; y++) {
	  for (unsigned int x=padding; x<bx; x++) {
	    target(x,y,z) = vol(x-padding,y-padding,z-padding);
	  }
	}
      }
      ColumnVector voxorig;
      voxorig = target.getvoxelorigin();
      voxorig += (float) padding;
      target.setvoxelorigin(voxorig);
      return 0;
    }


  int pad(volume& vol, int nx, int ny, int nz, int padding, float padval)
    {
      // resizes this volume, and pads it with padval...
      if ( (vol.slices()==nz) && (vol.rows()==ny) && (vol.columns()==nx) &&
	   (padding <= 0) )
	return 0;
      volume workingvol(nx,ny,nz);
      pad(vol,workingvol,padding,padval);
      vol.reinitialize(workingvol);
      return 0;
    }


  int unpad(const volume& vol, volume& target, int padding)
    {
      // copies the volume vol to target stripping the outer "padding" voxels
      unsigned int bx, by, bz;
      bx = vol.columns();
      by = vol.rows();
      bz = vol.slices(); 
      target.reinitialize(bx-2*padding,by-2*padding,bz-2*padding);
      for (unsigned int z=padding; z<bz-padding; z++) {
	for (unsigned int y=padding; y<by-padding; y++) {
	  for (unsigned int x=padding; x<bx-padding; x++) {
	    target(x-padding,y-padding,z-padding) = vol(x,y,z);
	  }
	}
      }
      ColumnVector voxorig;
      voxorig = target.getvoxelorigin();
      voxorig -= (float) padding;
      target.setvoxelorigin(voxorig);
      return 0;
    }



  int unpad(volume& vol, int padding)
    {
      if (padding <= 0) return 0;
      volume workingvol;
      unpad(vol,workingvol,padding);
      vol.reinitialize(workingvol);
      return 0;
    }


   int clamp(volume& vol, float minval, float maxval)
   {
     if (maxval < minval) return -1;
      for (int z=0; z<vol.zsize(); z++) {
	for (int y=0; y<vol.ysize(); y++) {
	  for (int x=0; x<vol.xsize(); x++) {
	    if (vol(x,y,z)>maxval)       vol(x,y,z)=maxval;
	    else if (vol(x,y,z)<minval)  vol(x,y,z)=minval;
	  }
	}
      }
      return 0;
   }
   

  int isotropic_rescale(volume& iso, const volume& aniso,
			float scale, bool verbose)
    {
      // takes the anisotropic volume, with given sampling and resamples
      // to an isotropic scale given by scale
      int sx, sy, sz, ix, iy, iz, x, y, z;
      float fx, fy, fz, stepx, stepy, stepz, bval = aniso.backgroundval();
      int bx=aniso.columns()-1, by=aniso.rows()-1, bz=aniso.slices()-1;
      stepx = scale / aniso.getx();
      stepy = scale / aniso.gety();
      stepz = scale / aniso.getz();
      sz = (int) (((float) aniso.slices() )  / stepz);
      sy = (int) (((float) aniso.rows() )    / stepy);
      sx = (int) (((float) aniso.columns() ) / stepx);
      iso.reinitialize(sx,sy,sz);
      iso.setvoxelsize(scale,scale,scale);
      for (fz=0.0, z=0; z<sz; z++, fz+=stepz) {
	for (fy=0.0, y=0; y<sy; y++, fy+=stepy) {
	  for (fx=0.0, x=0; x<sx; x++, fx+=stepx) {
	    ix = (int) fx;  iy = (int) fy;  iz = (int) fz;
	    if (quick_in_bounds(ix,iy,iz,bx,by,bz)) {
	      iso(x,y,z) = q_tri_interpolation(aniso,fx,fy,fz,ix,iy,iz);
	    } else {
	      iso(x,y,z) = bval;
	    }
	  }
	}
	if (verbose) cout << ".";
      }
      if (verbose) cout << endl;
      return 0;
    }


  ////////////////////////////////////////////////////////////////////////



  int make_separable_grad_masks(ColumnVector& blur, ColumnVector& grad)
    {
      blur.ReSize(3);
      grad.ReSize(3);
      grad <<  -1.0 << 0.0 << 1.0;
      // a Gaussian of sigma = 1
      blur << 0.2741 << 0.4519 << 0.2741;
      return 0;
    }


  int make_grad_masks(volume& maskx, volume& masky, volume& maskz)
    {
      maskx.reinitialize(3,3,3);
      masky.reinitialize(3,3,3);
      maskz.reinitialize(3,3,3);
      for (int z=0; z<3; z++) {
	for (int y=0; y<3; y++) {
	  for (int x=0; x<3; x++) {
	    maskx(x,y,z)=(x-1.0)*pow(3.0,1.0-fabs(y-1)-fabs(z-1));
	    masky(x,y,z)=(y-1.0)*pow(3.0,1.0-fabs(x-1)-fabs(z-1));
	    maskz(x,y,z)=(z-1.0)*pow(3.0,1.0-fabs(x-1)-fabs(y-1));
	  }
	}
      }
      return 0;
    }


  int make_blur_mask(ColumnVector& bmask, const float final_vox_dim, 
		     const float init_vox_dim)
    {
      if (fabs(init_vox_dim)<1e-8) { return -1; }
      float sampling_ratio = final_vox_dim / init_vox_dim;
      if (sampling_ratio < 1.1) {
	bmask.ReSize(1);
	bmask = 1.0;
	return 0;
      }
      float sigma = 0.85*(sampling_ratio/2.0);
      if (sigma<0.5) {
	bmask.ReSize(1);
	bmask = 1.0;
	return 0;
      }
      int n=((int) (sigma-0.001))*2 + 3;
      int midn = n/2 + 1;
      bmask.ReSize(n);
      for (int x=1; x<=n; x++) {
	bmask(x) = exp(-((float) Sqr(x-midn))/Sqr(sigma)/4.0);
      }
      bmask = bmask / Sum(bmask);
      return 0;
    }


  int convolve_indep(const volume& source, volume& target,
		     const ColumnVector& maskx, const ColumnVector& masky,
		     const ColumnVector& maskz) 
    {
      volume temp(source);
      target = source;
      target = 0.0;
      float val=0.0;
      int midx, midy, midz, midm=0;
      // if there is nothing to do - do nothing!
      if ((maskx.Nrows()==1) && (masky.Nrows()==1) && (maskz.Nrows()==1))
	{
	  if ( (fabs(maskx(1)-1.0)<1e-5) &&  (fabs(masky(1)-1.0)<1e-5)
	       && (fabs(maskz(1)-1.0)<1e-5) ) {
	    target = source;
	    return 0;
	  }
	}
      midx=maskx.Nrows()/2;
      midy=masky.Nrows()/2;
      midz=maskz.Nrows()/2;
      for (int dir=1; dir<=3; dir++) {
	switch (dir) 
	  {
	  case 1: midm=midx; val=maskx(1); break; 
	  case 2: midm=midy; val=masky(1); break; 
	  case 3: midm=midz; val=maskz(1); break; 
	  }
	if ((midm>1) || (fabs(val-1.0)>1e-5)) { 
	  for (int z=midz; z<source.slices()-midz; z++) {
	    for (int y=midy; y<source.rows()-midy; y++) {
	      for (int x=midx; x<source.columns()-midx; x++) {
		val = 0.0;
		for (int a=-midm; a<=midm; a++) {
		  switch (dir) 
		    {
		    case 1:
		      val+=temp(x+a,y,z) * maskx(a+midm+1);
		      break;
		    case 2:
		      val+=temp(x,y+a,z) * masky(a+midm+1);
		      break;
		    case 3:
		      val+=temp(x,y,z+a) * maskz(a+midm+1);
		      break;
		    }
		}
		target(x,y,z) = val;  // was   += val 
	      }
	    }
	    //cout << ".";
	  }
	} else {
	  target = temp;
	}
	if (dir<=2) { temp = target; target = 0.0; }
      }
      //cout << endl;
      // force the recalculation of the cog
      return 0;
    }


  int blur4subsampling(volume& target, const volume& source,
		       const ColumnVector& final_vox_dims)
    {
      //target = source;
      ColumnVector bmaskx, bmasky, bmaskz;
      make_blur_mask(bmaskx,final_vox_dims(1),source.getx());
      make_blur_mask(bmasky,final_vox_dims(2),source.gety());
      make_blur_mask(bmaskz,final_vox_dims(3),source.getz());
      //cout << "Blurring with masks :\n  " << bmaskx.t() << "  "
      //   << bmasky.t() << "  " << bmaskz.t();
      return convolve_indep(source,target,bmaskx,bmasky,bmaskz);
    }


  int blur4subsampling(volume& target, const volume& source,
		       float final_iso_vox_dims)
    {
      ColumnVector final_vox_dims(3);
      final_vox_dims = final_iso_vox_dims;
      return blur4subsampling(target,source,final_vox_dims);
    }


  int gradient_separably(const volume& source, volume& grad)
    {
      ColumnVector blurm,gradm;
      make_separable_grad_masks(blurm,gradm);
      float valx,valy,valz;;
      int midb, midg, midm;
      midb=blurm.Nrows()/2;
      midg=gradm.Nrows()/2;
      midm = Max(midb,midg);
      volume blurred;
      convolve_indep(source,blurred,blurm,blurm,blurm);
      for (int z=midg; z<source.slices()-midg; z++) {
	for (int y=midg; y<source.rows()-midg; y++) {
	  for (int x=midg; x<source.columns()-midg; x++) {
	    valx=0.0; valy=0.0; valz=0.0;
	    for (int a=-midg; a<=midg; a++) {
	      valx += blurred(x+a,y,z) * gradm(a+midg+1);
	      valy += blurred(x,y+a,z) * gradm(a+midg+1);
	      valz += blurred(x,y,z+a) * gradm(a+midg+1);
	    }
	    grad(x,y,z)=sqrt(Sqr(valx) + Sqr(valy) + Sqr(valz));
	  }
	}
	//cout << ".";
      }
      //cout << endl;
      return 0;
    }

  
  int gradient(const volume& source, volume& grad, bool verbose)
    {
      volume maskx,masky,maskz;
      make_grad_masks(maskx,masky,maskz);
      grad = source;
      grad = source.backgroundval();
      float valx, valy, valz;
      int midx, midy, midz;
      midz=maskx.slices()/2;
      midy=maskx.rows()/2;
      midx=maskx.columns()/2;
      for (int z=midz; z<source.slices()-midz; z++) {
	for (int y=midy; y<source.rows()-midy; y++) {
	  for (int x=midx; x<source.columns()-midx; x++) {
	    valx=0.0; valy=0.0; valz=0.0;
	    for (int mz=-midz; mz<=midz; mz++) {
	      for (int my=-midy; my<=midy; my++) {
		for (int mx=-midx; mx<=midx; mx++) {
		  valx+=source(x+mx,y+my,z+mz) * maskx(mx+midx,my+midy,mz+midz);
		  valy+=source(x+mx,y+my,z+mz) * masky(mx+midx,my+midy,mz+midz);
		  valz+=source(x+mx,y+my,z+mz) * maskz(mx+midx,my+midy,mz+midz);
		}
	      }
	    }
	    grad(x,y,z)=sqrt(Sqr(valx) + Sqr(valy) + Sqr(valz));
	  }
	}
	if (verbose) cout << ".";
      }
      if (verbose) cout << endl;
      return 0;

    }

  /*
  int aniso_gradient(const volume& source, volume& grad, bool verbose)
    {
      volume maskx,masky,maskz;
      make_grad_masks(maskx,masky,maskz);
      grad = source;
      grad = source.backgroundval();
      float valx, valy, valz;
      float xsize, ysize, zsize;
      int midx, midy, midz;
      midz=maskx.slices()/2;
      midy=maskx.rows()/2;
      midx=maskx.columns()/2;
      xsize = source. getx();
      ysize = source. gety();
      zsize = source. getz();
      for (int z=midz; z<source.slices()-midz; z++) {
	for (int y=midy; y<source.rows()-midy; y++) {
	  for (int x=midx; x<source.columns()-midx; x++) {
	    valx=0.0; valy=0.0; valz=0.0;
	    for (int mz=-midz; mz<=midz; mz+=zsize) {
	      for (int my=-midy; my<=midy; my+=ysize) {
		for (int mx=-midx; mx<=midx; mx+=xsize) {
		  valx+=source(x+mx,y+my,z+mz) * maskx(mx+midx,my+midy,mz+midz);
		  valy+=source(x+mx,y+my,z+mz) * masky(mx+midx,my+midy,mz+midz);
		  valz+=source(x+mx,y+my,z+mz) * maskz(mx+midx,my+midy,mz+midz);
		}
	      }
	    }
	    grad(x,y,z)=sqrt(Sqr(valx) + Sqr(valy) + Sqr(valz));
	  }
	}
	if (verbose) cout << ".";
      }
      if (verbose) cout << endl;
      return 0;

    }
  */

  volume gaussian_kernel(float sigma, int radius)
    {
      volume new_kernel((2*radius+1),(2*radius+1),1); 
      float sum;
      sum=0.0f;
      
      for(int i=-radius;i<=radius;i++) {
	for(int j=-radius;j<=radius;j++)
	  {
	    new_kernel((j+radius),(i+radius),0) = 
	      pow(2.71, -(i*i+j*j)/(2*sigma*sigma));
	    sum+=new_kernel((j+radius), (i+radius), 0);
	  }
      }
      
      for(int i=-radius;i<=radius;i++) {
	for(int j=-radius;j<=radius;j++) {
	  new_kernel((j+radius),(i+radius),0) /= sum;
	}
      }

      return new_kernel;
    }


  int convolve(const volume& source, volume& result, const volume& mask,
	       bool verbose)
    {
      result = source;
      result = source.backgroundval();
      float val;
      if ( ((mask.slices() % 2)==0) || ((mask.rows() % 2)==0) 
	   || ((mask.columns() % 2)==0) ) {
	cerr << "Cannot deal with even dimensions in mask\n";
	return -1;
      }
      int midx, midy, midz;
      midz=mask.slices()/2;
      midy=mask.rows()/2;
      midx=mask.columns()/2;
      for (int z=midz; z<source.slices()-midz; z++) {
	for (int y=midy; y<source.rows()-midy; y++) {
	  for (int x=midx; x<source.columns()-midx; x++) {
	    val=0.0;
	    for (int mz=-midz; mz<=midz; mz++) {
	      for (int my=-midy; my<=midy; my++) {
		for (int mx=-midx; mx<=midx; mx++) {
		  val+=source(x+mx,y+my,z+mz) * mask(mx+midx,my+midy,mz+midz);
		}
	      }
	    }
	    result(x,y,z)=val;
	  }
	}
	if (verbose) cout << ".";
      }
      if (verbose) cout << endl;
      return 0;
    }


  int mirror_convolve(const volume& source, volume& result, const volume& mask,
	       bool verbose)
    {
      result = source;
      result = source.backgroundval();
      float val;
      if ( ((mask.slices() % 2)==0) || ((mask.rows() % 2)==0) 
	   || ((mask.columns() % 2)==0) ) {
	cerr << "Cannot deal with even dimensions in mask\n";
	return -1;
      }
      int midx, midy, midz;
      int xloc, yloc, zloc;
      midz=mask.slices()/2;
      midy=mask.rows()/2;
      midx=mask.columns()/2;
      
      for (int z=0; z<source.slices(); z++) {
	for (int y=0; y<source.rows(); y++) {
	  for (int x=0; x<source.columns(); x++) {
	    val=0.0;
	    for (int mz=-midz; mz<=midz; mz++) {
	      for (int my=-midy; my<=midy; my++) {
		for (int mx=-midx; mx<=midx; mx++) {
		  xloc = x+mx; yloc = y+my; zloc = z+mz;
     		  while (!source.in_bounds(xloc, yloc, zloc)) {
		    if (xloc >= source.columns())
		      xloc = 2*source.columns() - xloc - 1;
		    else if (xloc < 0)
		      xloc = -xloc;
		    if (yloc >= source.rows())
		      yloc = 2*source.rows() - yloc - 1;
		    else if (yloc < 0)
		      yloc = -yloc;
		    if (zloc >= source.slices())
		      zloc = 2*source.slices() - zloc - 1;
		    else if (zloc < 0)
		      zloc = -zloc;
		  }
		  val+=source(xloc,yloc,zloc) * mask(mx+midx,my+midy,mz+midz);
		}
	      }
	    }
	    result(x,y,z)=val;
	  }
	}
	if (verbose) cout << ".";
      }
      if (verbose) cout << endl;
      return 0;
    }


  /* detects closed contour edges in a volume as the zero crossings of the 
     Laplacian of a Gaussian (LoG). This is implemented by convolving the 
     data with a kernel formed by subtracting a Gaussian of radius sigma1
     from a second Gaussian kernel of radius sigma2 (sigma1 < sigma2) */

  int log_edge_detect(volume& source, float sigma1, float sigma2)
    {
      float zero_tolerance = 1000.0; /* what we define as an acceptable "zero-crossing" */
      int radius1;
      volume log_kern, temp_kern;
      volume result = source;

      radius1 = (int)(4*sigma2);
      log_kern = gaussian_kernel(sigma2, radius1);
      temp_kern = gaussian_kernel(sigma1, radius1);
      
      for (int i = 0; i <= 2*radius1; i++) {
	for (int j = 0; j <= 2*radius1; j++) {
	  log_kern(i,j,0) -= temp_kern(i,j,0);
	}
      }

      mirror_convolve(source, result, log_kern);

      for (int z=0; z<result.slices(); z++) {
	for (int y=0; y<result.rows(); y++) {
	  for (int x=0; x<result.columns(); x++) {
	    if ( fabs(result(x,y,z)) < zero_tolerance)
	      source(x,y,z) = 0.0;
	    else
	      source(x,y,z) = 1.0;
	  }
	}
      }
	      
      return 0;
    }

   int fixed_edge_detect(volume& source)
   {
     fixed_edge_detect(source, 10000.0);
     
     return 0;
   }

   int fixed_edge_detect(volume& source, float toler)
   {
     volume log_kern(3,3,1);
     volume result = source;

     for (int i = 0; i <= 2; i++) {
       for (int j = 0; j <= 2; j++) {
	 log_kern(i,j,0) = -1;
       }
     }
     log_kern(1,1,0) = 8;

     /*
     for (int i = 0; i <= 2; i++) {
       for (int j = 0; j <= 2; j++) {
	 if (i == 0 || j == 0 || i == 2 || j == 2)
	   log_kern(i,j,0) = 0;
	 else 
	   log_kern(i,j,0) = -1;
       }
     }
     log_kern(1,1,0) = 4;
     */

     /*
     for (int i = 0; i <= 2; i++) {
       for (int j = 0; j <= 2; j++) {
	 if (i == 0 || j == 0 || i == 2 || j == 2)
	   log_kern(i,j,0) = 1;
	 else 
	   log_kern(i,j,0) = -2;
       }
     }
     log_kern(1,1,0) = 4;
     */

     mirror_convolve(source, result, log_kern);
     
     for (int z=0; z<result.slices(); z++) {
       for (int y=0; y<result.rows(); y++) {
	 for (int x=0; x<result.columns(); x++) {
	   if ( fabs(result(x,y,z)) < toler)
	     source(x,y,z) = 0.0;
	   else
	     source(x,y,z) = 1.0;
	 }
       }
     }
	      
      return 0;
   }

  int subsample_by2(const volume& refvol, volume& halfvol)
    {
      // subsamples a volume (refvol) by blurring and subsampling to give
      //  a new volume (halfvol)
      // note that this creates the new volume, throwing away any data
      //  previous contained therein
      halfvol.reinitialize(refvol.columns()/2,refvol.rows()/2,refvol.slices()/2);
      halfvol = refvol.backgroundval();
      halfvol.setvoxelsize(refvol.getx() * 2.0, refvol.gety() * 2.0,
		   refvol.getz() * 2.0);
      halfvol.setvoxelorigin(refvol.getvoxelorigin()/2.0);
      for (int z=1, bz=2; bz<refvol.slices()-1; z++,bz+=2) {
	for (int y=1, by=2; by<refvol.rows()-1; y++,by+=2) {
	  for (int x=1, bx=2; bx<refvol.columns()-1; x++,bx+=2) {
	    halfvol(x,y,z) = 0.0923 * (refvol(bx,by,bz))
	      + 0.0560 * (refvol(bx+1,by,bz) + 
			  refvol(bx-1,by,bz) +
			  refvol(bx,by+1,bz) + 
			  refvol(bx,by-1,bz) +
			  refvol(bx,by,bz+1) + 
			  refvol(bx,by,bz-1))
	      + 0.0339 * (refvol(bx+1,by+1,bz) + 
			  refvol(bx+1,by-1,bz) +
			  refvol(bx-1,by+1,bz) + 
			  refvol(bx-1,by-1,bz) +
			  refvol(bx+1,by,bz+1) + 
			  refvol(bx+1,by,bz-1) +
			  refvol(bx-1,by,bz+1) + 
			  refvol(bx-1,by,bz-1) +
			  refvol(bx,by+1,bz+1) + 
			  refvol(bx,by+1,bz-1) +
			  refvol(bx,by-1,bz+1) + 
			  refvol(bx,by-1,bz-1))
	      + 0.0206 * (refvol(bx+1,by+1,bz+1) + 
			  refvol(bx+1,by+1,bz-1) +
			  refvol(bx+1,by-1,bz+1) + 
			  refvol(bx+1,by-1,bz-1) +
			  refvol(bx-1,by+1,bz+1) + 
			  refvol(bx-1,by+1,bz-1) +
			  refvol(bx-1,by-1,bz+1) + 
			  refvol(bx-1,by-1,bz-1));
	  }
	}
      }
      return 0;
    }


  //////////////////////////////////////////////////////////////////////////
 
  // AFFINE TRANSFORM

  int affine_transform(volume& vout, const volume& vin, const Matrix& aff)
    {
      // takes the volume (vin) and applies a spatial transform, given
      //  by the 4x4 matrix (aff) in WORLD COORDINATES
      // the result is a new volume (vout)


      // Do everything in practice via the inverse transformation
      // That is, for every point in vout, calculate the pre-image in
      //  vin to which it corresponds, and interpolate vin to get the
      //  value there.
      // Also, the sampling transformations must be accounted for:
      //     T_vox1->vox2 = (T_samp2)^-1 * T_world * T_samp1
      Matrix iaffbig = vin.sampling_matrix().i() * aff.i() *
	                     vout.sampling_matrix();  
      Matrix iaff=iaffbig.SubMatrix(1,3,1,3);
      unsigned int xb1=vin.columns()-1, yb1=vin.rows()-1, zb1=vin.slices()-1;
      unsigned int xb2=vout.columns()-1, yb2=vout.rows()-1, zb2=vout.slices()-1;
      int io1, io2, io3;
      float bval = vin.backgroundval();

      float a11=iaff(1,1), a12=iaff(1,2), a13=iaff(1,3), a14=iaffbig(1,4),
	a21=iaff(2,1), a22=iaff(2,2), a23=iaff(2,3), a24=iaffbig(2,4),
	a31=iaff(3,1), a32=iaff(3,2), a33=iaff(3,3), a34=iaffbig(3,4), o1,o2,o3;
  
  // cout << "Bounds are " << xb << " " << yb << " " << zb << endl;
  // The matrix algebra below has been hand-optimized from
  //  [o1 o2 o3] = a * [x y z]  at each iteration
      for (unsigned int z=0; z<=zb2; z++) { 
	for (unsigned int x=0; x<=xb2; x++) { 
	  o1=x*a11 + z*a13 + a14;  // y=0
	  o2=x*a21 + z*a23 + a24;  // y=0
	  o3=x*a31 + z*a33 + a34;  // y=0
	  for (unsigned int y=0; y<=yb2; y++) {
	    if ( (o1>0.0) && (o2>0.0) && (o3>0.0) ) {
	      io1=(int) o1;
	      io2=(int) o2;
	      io3=(int) o3;
	      //      newpos(1)=x; newpos(2)=y; newpos(3)=z;
	      //      oldpos = iaff * newpos + itrans;
	      if (quick_max_bounds(io1, io2, io3, xb1, yb1, zb1) )
		{
		  vout(x,y,z) = 
		    q_tri_interpolation(vin,o1,o2,o3,io1,io2,io3);
		} else {
		  vout(x,y,z) = bval;
		}
	    } else {
	      vout(x,y,z) = bval;
	    }
	    o1 += a12;
	    o2 += a22;
	    o3 += a32;
	  }
	}
	//cout << ".";
      }
      return 0;
    }



  int filled_affine_transform(volume& vout, const volume& vin, 
			      const Matrix& aff)
    {
      // does the transform with an effective one voxel padding of
      // vin around the edges (achieves this without changing or
      // copying vin)

      // takes the volume (vin) and applies a spatial transform, given
      //  by the 4x4 matrix (aff) in WORLD COORDINATES
      // the result is a new volume (vout)


      // Do everything in practice via the inverse transformation
      // That is, for every point in vout, calculate the pre-image in
      //  vin to which it corresponds, and interpolate vin to get the
      //  value there.
      // Also, the sampling transformations must be accounted for:
      //     T_vox1->vox2 = (T_samp2)^-1 * T_world * T_samp1
      Matrix iaffbig = vin.sampling_matrix().i() * aff.i() *
	                     vout.sampling_matrix();  
      Matrix iaff=iaffbig.SubMatrix(1,3,1,3);
      int xb1=vin.columns()-1, yb1=vin.rows()-1, zb1=vin.slices()-1;
      unsigned int xb2=vout.columns()-1, yb2=vout.rows()-1, zb2=vout.slices()-1;
      int io1, io2, io3;
      float bval = vin.backgroundval();

      float a11=iaff(1,1), a12=iaff(1,2), a13=iaff(1,3), a14=iaffbig(1,4),
	a21=iaff(2,1), a22=iaff(2,2), a23=iaff(2,3), a24=iaffbig(2,4),
	a31=iaff(3,1), a32=iaff(3,2), a33=iaff(3,3), a34=iaffbig(3,4), o1,o2,o3;
  
  // cout << "Bounds are " << xb << " " << yb << " " << zb << endl;
  // The matrix algebra below has been hand-optimized from
  //  [o1 o2 o3] = a * [x y z]  at each iteration
      for (unsigned int z=0; z<=zb2; z++) { 
	for (unsigned int x=0; x<=xb2; x++) { 
	  o1=x*a11 + z*a13 + a14;  // y=0
	  o2=x*a21 + z*a23 + a24;  // y=0
	  o3=x*a31 + z*a33 + a34;  // y=0
	  for (unsigned int y=0; y<=yb2; y++) {
//  	    io1=(int) floor(o1);
//  	    io2=(int) floor(o2);
//  	    io3=(int) floor(o3);
//  	    if (filled_bounds(io1, io2, io3, xb1, yb1, zb1) )
//  	      {
		vout(x,y,z) = 
		  filled_tri_interpolation(vin,o1,o2,o3);
//  	      } else {
//  		vout(x,y,z) = bval;
//  	      }
	    o1 += a12;
	    o2 += a22;
	    o3 += a32;
	  }
	}
	//cout << ".";
      }
      return 0;
    }



  int affine_transform(imagepair* ims, const Matrix& aff)
    {
      return affine_transform(ims->outputvol,ims->testvol,aff);
    }


  /////////////////////////////////////////////////////////////////////////////

  long int no_voxels(const volume& vol) {
    return (long int) (vol.rows() * vol.columns() * vol.slices());
  }

  void imagepair::set_no_bins(int n_bins) 
    {
      no_bins=n_bins;
      //hist.ReSize(no_bins+1,no_bins+1);
      jointhist = new int[(no_bins+1)*(no_bins+1)];
      marghist1 = new int[no_bins+1];
      marghist2 = new int[no_bins+1];
      fjointhist = new float[(no_bins+1)*(no_bins+1)];
      fmarghist1 = new float[no_bins+1];
      fmarghist2 = new float[no_bins+1];
      int N = no_voxels(this->refvol);
      float p=0.0;
      plnp.ReSize(Min((int) 10000, (int) (10*N/(no_bins+1))));
      for (int num=1; num<=plnp.Nrows(); num++) {
	p = ((float) num) / ((float) N);
	plnp(num) = -p*log(p);
      }

      if (bindex) delete [] bindex;
      bindex = new int[refvol.columns()*refvol.rows()*refvol.slices()];
      float a1=no_bins/(refmax-refmin);
      float a0=-refmin*no_bins/(refmax-refmin);
      int *bptr = bindex;
      for (int z=0; z<refvol.zsize(); z++) { 
	for (int y=0; y<refvol.ysize(); y++) { 
	  for (int x=0; x<refvol.xsize(); x++) {
	    *bptr = (int) (refvol(x,y,z)*a1 + a0);
	    if (*bptr >= no_bins)  *bptr = no_bins-1;
	    if (*bptr < 0.0)  *bptr = 0;
	    bptr++;
	  }
	}
      }
    }


  /////////////////////////////////////////////////////////////////////////////

   int calc_histogram(const volume& vol, int nbins, float minval, float maxval, 
		      ColumnVector& hist)
   {
     if (hist.Nrows()!=nbins) hist.ReSize(nbins);
     hist = 0.0;

     if (maxval <= minval) return -1;

     float a=((float) nbins)/(maxval - minval);
     float b= - ((float) nbins)*minval/(maxval - minval);
     int binno = 0;

     for (int z=0; z<vol.zsize(); z++) { 
       for (int y=0; y<vol.ysize(); y++) { 
	 for (int x=0; x<vol.xsize(); x++) {
	   binno = (int) (a*vol(x,y,z) + b);
	   if (binno > nbins-1)  binno=nbins-1;
	   if (binno < 0)      binno=0;
	   hist(binno+1)++;
	 }
       }
     }
     return 0; 
   }


   int find2and98percentiles(const ColumnVector& hist, float& f2, float& f98,
			     float minval, float maxval)
   {
     // Returns bin2 and bin98 such that sun from bin2 to bin98 INCLUSIVE
     //  is guaranteed to have the inner 96%   (at least)
     // f2 is the lower bound of bin2, and f98 is the upper bound of bin98
     // Note that all bin indices go from 0 to nbins-1  ( +1 needed for hist() )
     float count=0;
     int bin=0, bin2=0, bin98=0, nbins = hist.Nrows();
     float sumhist = hist.Sum();
     if (sumhist <= 1) { cerr << "Histogram is null" << endl; return -1; }
     while (count<sumhist*0.02) {
       count += hist(bin+1);
       bin++;
     }
     // last value caused count to exceed threshold, but since then bin++
     bin2 = bin - 1;
     f2 = minval + bin2 * (maxval - minval)/((float) nbins);

     count = 0;
     bin = nbins-1;
     while (count<sumhist*0.02) {
       count += hist(bin+1);
       bin--;
     }
     // last value caused count to exceed threshold, but since then bin--
     bin98 = bin + 1;
     f98 = minval + (bin98 + 1) * (maxval - minval)/((float) nbins);
     
     return 0;
   }


   int find_robust_limits(const volume& vol, int nbins, ColumnVector& hist,
			  float& minval, float& maxval)
   {
     if (hist.Nrows()!=nbins) hist.ReSize(nbins);

     float absmin, absmax, f2, f98;
     get_min_max(vol,absmin, absmax);

     minval = absmin;  maxval = absmax;
     calc_histogram(vol,nbins,minval,maxval,hist);
     find2and98percentiles(hist,f2,f98,minval,maxval);

//       if ((f98 - f2) > (maxval - minval) * 0.1) {
//         return 0;
//       }

//       // if this doesn't work, get the true 2 and 98 percentiles and use these
//       ColumnVector plimits(2), flimits;
//       plimits << 2 << 98;
//       flimits = get_percentiles(vol,plimits);
//       minval = flimits(1);
//       maxval = flimits(2);
//       calc_histogram(vol,nbins,minval,maxval,hist);
//       // check that this is OK
//       find2and98percentiles(hist,f2,f98,minval,maxval);
//       if ((f98 - f2) > (maxval - minval) * 0.1) {
//         return 0;
//       }

//       // else it is not OK, so revert to the absolute values
//       cerr << "WARNING: had difficulty finding robust limits in histogram\n";
//       minval = absmin;  maxval = absmax;
//       calc_histogram(vol,nbins,minval,maxval,hist);

//       return 1;
//     }
     

     for (int passes=1; passes<10; passes++) {
       if ((f98 - f2) > (maxval - minval) * 0.1) {
	 // OK if the middle 96% of data is more than 10% of histogram range
	 return 0;
       }
       // Otherwise start with the 2 and 98 percentile limits and try again
       minval = f2;  maxval = f98;
       calc_histogram(vol,nbins,minval,maxval,hist);
       find2and98percentiles(hist,f2,f98,minval,maxval);
     }

     // exceeded 10 iterations, so use absolute range but without end bins
     cerr << "WARNING: had difficulty finding robust limits in histogram\n";
     minval = absmin;  maxval = absmax;
     calc_histogram(vol,nbins,minval,maxval,hist);
     // strip off the outer bins to create a new histogram
     ColumnVector limitedhist(hist.Nrows()-2);
     for (int n=1; n<=limitedhist.Nrows(); n++) {
       limitedhist(n) = hist(n+1);
     }
     float binwidth = (maxval - minval)/((float) nbins);
     find2and98percentiles(hist,f2,f98,minval+binwidth,maxval-binwidth);

     return 1;
   }


#ifndef NO_NAMESPACE
 }
#endif
