// General IO functions (images and transformation files)
//  Written by Mark Jenkinson  2000

#if !defined(__newimageio_h)
#define __newimageio_h

#include <string>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <strstream>
#include "newimage.h"
#include "newmatio.h"
#include "avwio.h"
#include "miscmaths.h"

using namespace NEWMAT;

namespace NEWIMAGE {

typedef AVW volumeinfo;

int make_basename(string& filename);

template <class T>
int read_volume(volume<T>& target, const string& filename);
template <class T>
int read_volume(volume<T>& target, const string& filename, volumeinfo& vinfo);
template <class T>
int read_volume_hdr_only(volume<T>& target, const string& filename);
template <class T>
int read_volume_hdr_only(volume<T>& target, const string& filename, 
			 volumeinfo& vinfo);
int read_complexvolume(volume<float>& realvol, volume<float>& imagvol,
		       const string& filename);

template <class T>
int read_volume4D(volume4D<T>& target, const string& filename);
template <class T>
int read_volume4D(volume4D<T>& target, const string& filename, 
		  volumeinfo& vinfo);
template <class T>
int read_volume4D_hdr_only(volume4D<T>& target, const string& filename);
template <class T>
int read_volume4D_hdr_only(volume4D<T>& target, const string& filename,
			   volumeinfo& vinfo);
int read_complexvolume4D(volume4D<float>& realvol, volume4D<float>& imagvol,
			 const string& filename);



template <class T>
int save_volume(const volume<T>& source, const string& filename);
template <class T>
int save_volume(const volume<T>& source, const string& filename,
		const volumeinfo& vinfo);
int save_complexvolume(const volume<float>& realvol, 
		       const volume<float>& imagvol, const string& filename);
int save_complexvolume(const volume<float>& realvol, 
		       const volume<float>& imagvol, const string& filename,
		       const volumeinfo& vinfo);

template <class T>
int save_volume4D(const volume4D<T>& source, const string& filename);
template <class T>
int save_volume4D(const volume4D<T>& source, const string& filename,
		  const volumeinfo& vinfo);
int save_complexvolume4D(const volume4D<float>& realvol, 
			 const volume4D<float>& imagvol, 
			 const string& filename);
int save_complexvolume4D(const volume4D<float>& realvol, 
			 const volume4D<float>& imagvol, 
			 const string& filename,
			 const volumeinfo& vinfo);


  // Transform matrix IO
  
int read_ascii_matrix(Matrix &target, const string& filename);
  //int write_ascii_matrix(const Matrix& affmat, const string& filename);
int get_outputusermat(const string& filename, Matrix& oumat);

template <class T>
int read_matrix(Matrix &target, const string& filename, const volume<T>& invol);
template <class T>
int read_matrix(Matrix &target, const string& filename, const volume<T>& invol,
		const volume<T>& finalvol);
template <class T>
int write_medx_matrix(const Matrix& worldmat, const string& filename, 
		      const volume<T>& initvol, 
		      const volume<T>& finalvol, 
		      const string& mtype, const string& reffname);

// Helper functions

short dtype(const volume<char>& vol);
short dtype(const volume<short>& vol);
short dtype(const volume<int>& vol);
short dtype(const volume<float>& vol);
short dtype(const volume<double>& vol);

short dtype(const volume4D<char>& vol);
short dtype(const volume4D<short>& vol);
short dtype(const volume4D<int>& vol);
short dtype(const volume4D<float>& vol);
short dtype(const volume4D<double>& vol);

short dtype(const string& filename);

volumeinfo volinfo(const string& filename);

////////////////////////////////////////////////////////////////////////
///////////////////////// TEMPLATE DEFINITIONS /////////////////////////
////////////////////////////////////////////////////////////////////////

template <class T>
void AvwReadBuffer(AVW* IP, T* tbuffer) 
{
  short sx,sy,sz,st;
  AvwGetDim(IP,&sx,&sy,&sz,&st);
  size_t imagesize=sx*sy*sz;
  short type;
  AvwGetDataType(IP,&type);
  switch(type)
    {
    case DT_SIGNED_SHORT:
      {
	short* sbuffer=new short[imagesize];
	AvwReadVolumes(IP,sbuffer,1);
	convertbuffer(sbuffer,tbuffer,imagesize);
	delete[] sbuffer;
      }
      break;
    case DT_UNSIGNED_CHAR:
      {
	unsigned char* sbuffer=new unsigned char[imagesize];
	AvwReadVolumes(IP,sbuffer,1);
	convertbuffer(sbuffer,tbuffer,imagesize);
	delete[] sbuffer;
      }
      break;
    case DT_SIGNED_INT:
      {
	int* sbuffer=new int[imagesize];
	AvwReadVolumes(IP,sbuffer,1);
	convertbuffer(sbuffer,tbuffer,imagesize);
	delete[] sbuffer;
      }
      break;
    case DT_FLOAT:
      {
	float* sbuffer=new float[imagesize];
	AvwReadVolumes(IP,sbuffer,1);
	convertbuffer(sbuffer,tbuffer,imagesize);
	delete[] sbuffer;
      }
      break;
    case DT_DOUBLE:
      {
	double* sbuffer=new double[imagesize];
	AvwReadVolumes(IP,sbuffer,1);
	convertbuffer(sbuffer,tbuffer,imagesize);
	delete[] sbuffer;
      }
      break;
    default:
      ostrstream errmsg;
      errmsg << "Avwread: DT " << type <<  " not supported";
      perror(errmsg.str());
    }
}
  


template <class T>
void load_avw_origin(volume<T>& vol, AVW* IP1) {
  short orig[5];
  AvwGetOriginator(IP1,orig);
  ColumnVector origin(3);
  origin(1) = orig[0];
  origin(2) = orig[1];
  origin(3) = orig[2];
  
  vol.setorigin(origin);
}


template <class T>
void save_avw_origin(const volume<T>& vol, AVW* IP1) {
  short orig[5];
  orig[0] = (short) vol.getorigin()(1);
  orig[1] = (short) vol.getorigin()(2);
  orig[2] = (short) vol.getorigin()(3);
  orig[3] = 0;
  orig[4] = 0;

  AvwSetOriginator(IP1,orig);
}



// External functions

// READ FUNCTIONS


void check_filename(const string& basename);

template <class T>
int read_volume(volume<T>& target, const string& filename, 
		volumeinfo& vinfo, short& dtype, bool read_img_data)
{
  Tracer trcr("read_volume");
  if ( filename.size()<1 ) return -1;
  string basename = filename;
  make_basename(basename);
  check_filename(basename);

  AVW* IP1 = AvwOpen(basename.c_str(), "r");
  short sx,sy,sz,st;
  AvwGetDim(IP1,&sx,&sy,&sz,&st);
  size_t volsize=sx*sy*sz;

  T* tbuffer=new T[volsize];
  if (read_img_data)  AvwReadBuffer(IP1,tbuffer);
  target.reinitialize(sx,sy,sz,tbuffer,true);

  float x,y,z,tr;
  AvwGetVoxDim(IP1,&x,&y,&z,&tr);
  target.setdims(x,y,z);

  AvwGetDataType(IP1,&dtype);
  load_avw_origin(target,IP1);

  AvwCloneHeader(&vinfo,IP1);
  AvwClose(IP1);
  return 0;
}


template <class T>
int read_volume(volume<T>& target, const string& filename, volumeinfo& vinfo)
{
  short dtype;
  int retval = read_volume(target,filename,vinfo,dtype,true);
  return retval;
}

template <class T>
int read_volume(volume<T>& target, const string& filename)
{
  short dtype;
  volumeinfo vinfo;
  int retval = read_volume(target,filename,vinfo,dtype,true);
  return retval;
}

template <class T>
int read_volume_hdr_only(volume<T>& target, const string& filename, 
			 volumeinfo& vinfo)
{
  short dtype;
  int retval = read_volume(target,filename,vinfo,dtype,false);
  return retval;
}

template <class T>
int read_volume_hdr_only(volume<T>& target, const string& filename)
{
  short dtype;
  volumeinfo vinfo;
  int retval = read_volume(target,filename,vinfo,dtype,false);
  return retval;
}


template <class T>
int read_volume4D(volume4D<T>& target, const string& filename, 
		  volumeinfo& vinfo, short& dtype, bool read_img_data)
{
  Tracer trcr("read_volume4D");
  if ( filename.size()<1 ) return -1;
  string basename = filename;
  make_basename(basename);
  check_filename(basename);

  AVW* IP1 = AvwOpen(basename.c_str(), "r");
  short sx,sy,sz,st;
  AvwGetDim(IP1,&sx,&sy,&sz,&st);
  size_t volsize=sx*sy*sz;

  volume<T> dummyvol(sx,sy,sz);
  for (int t=0; t<st; t++) {
    target.addvolume(dummyvol);  // MJ NOTE : SHOULD CLEAR TARGET FIRST!!
    T* tbuffer=new T[volsize];
    if (read_img_data)  AvwReadBuffer(IP1,tbuffer);
    // Note that the d_owner flag = true in the following so that the
    //  control for delete is passed to the volume class
    target[t].reinitialize(sx,sy,sz,tbuffer,true);
  }

  float x,y,z,tr;
  AvwGetVoxDim(IP1,&x,&y,&z,&tr);
  target.setdims(x,y,z,tr);

  AvwGetDataType(IP1,&dtype);
  for (int t=0; t<target.tsize(); t++) {
    load_avw_origin(target[t],IP1);
  }

  AvwCloneHeader(&vinfo,IP1);
  AvwClose(IP1);
  return 0;
}


template <class T>
int read_volume4D(volume4D<T>& target, const string& filename, 
		  volumeinfo& vinfo)
{
  short dtype;
  int retval = read_volume4D(target,filename,vinfo,dtype,true);
  return retval;
}

template <class T>
int read_volume4D(volume4D<T>& target, const string& filename)
{
  short dtype;
  volumeinfo vinfo;
  int retval = read_volume4D(target,filename,vinfo,dtype,true);
  return retval;
}

template <class T>
int read_volume4D_hdr_only(volume4D<T>& target, const string& filename,
			   volumeinfo& vinfo)
{
  short dtype;
  int retval = read_volume4D(target,filename,vinfo,dtype,false);
  return retval;
}

template <class T>
int read_volume4D_hdr_only(volume4D<T>& target, const string& filename)
{
  short dtype;
  volumeinfo vinfo;
  int retval = read_volume4D(target,filename,vinfo,dtype,false);
  return retval;
}


// SAVE FUNCTIONS

template <class T>
int save_volume(const volume<T>& source, const string& filename, 
		const volumeinfo& vinfo, bool use_vinfo)
{
  Tracer tr("save_volume");
  string basename = filename;
  make_basename(basename);
  if ( basename.size()<1 ) return -1;

  AVW *OP=AvwOpen(basename.c_str(),"w");
  if (OP==0) return -1;
  if (use_vinfo) AvwCloneHeader(OP,&vinfo);
  
  int imin, imax;
  imin = (int) source.min();
  imax = (int) source.max();
  if (imax<=imin)  imax=imin+1;
  AvwSetMinMax(OP,imin,imax);
  save_avw_origin(source,OP);
    
  AvwSetDim(OP,source.xsize(),source.ysize(),source.zsize(),1);
  AvwSetDataType(OP, dtype(source));
  AvwSetVoxDim(OP,source.xdim() * source.xsign(),
	       source.ydim() * source.ysign(),
	       source.zdim() * source.zsign(),1.0);

  AvwWriteVolumes(OP,&(source(0,0,0)),1);

  AvwClose(OP); 
  return 0;
}


template <class T>
int save_volume(const volume<T>& source, const string& filename)
{
  volumeinfo vinfo;
  return save_volume(source,filename,vinfo,false);
}


template <class T>
int save_volume(const volume<T>& source, const string& filename,
		const volumeinfo& vinfo)
{
  return save_volume(source,filename,vinfo,true);
}



template <class T>
int save_volume4D(const volume4D<T>& source, const string& filename,
		  const volumeinfo& vinfo, bool use_vinfo)
{
  Tracer tr("save_volume4D");

  if (source.tsize()<=0) return -1;

  string basename = filename;
  make_basename(basename);
  if ( basename.size()<1 ) return -1;

  AVW* OP=AvwOpen(basename.c_str(),"w");
  if (OP==0) return -1;
  if (use_vinfo) AvwCloneHeader(OP,&vinfo);
    
  int imin, imax;
  imin = (int) source.min();
  imax = (int) source.max();
  if (imax<=imin)  imax=imin+1;
  AvwSetMinMax(OP,imin,imax);
  save_avw_origin(source[0],OP);
    
  AvwSetDim(OP,source.xsize(),source.ysize(),source.zsize(),source.tsize());
  AvwSetDataType(OP, dtype(source));
  AvwSetVoxDim(OP,source.xdim() * source.xsign(),
	       source.ydim() * source.ysign(),
	       source.zdim() * source.zsign(),
	       source.tdim());
  
  for (int t=0; t<source.tsize(); t++) {
    AvwWriteVolumes(OP,&(source[t](0,0,0)),1);
  }

  AvwClose(OP); 
  return 0;
}

template <class T>
int save_volume4D(const volume4D<T>& source, const string& filename)
{
  volumeinfo vinfo;
  return save_volume4D(source,filename,vinfo,false);
}


template <class T>
int save_volume4D(const volume4D<T>& source, const string& filename,
		  const volumeinfo& vinfo)
{
  return save_volume4D(source,filename,vinfo,true);
}

// MATRIX IO

// Non-templated helper functions (bodies in .cc)

int read_ascii_matrix(Matrix &target, ifstream& matfile);
int get_medx_small_matrix(Matrix &target, ifstream& matfile);
int get_medx_matrix(Matrix &target, ifstream& matfile);
int get_minc_matrix(Matrix &target, ifstream& matfile);
int put_medx_matrix(ofstream& matfile, const string& name, 
		    const Matrix& affmat);

// Template definitions

template <class T>
int medx2world(Matrix &target, const Matrix& medxmat, 
	       const ColumnVector& finalimsize, const Matrix& finalvoxsize,
	       const volume<T>& initvol) 
{
  Tracer tr("medx2world");
  // convert the voxel->voxel MEDx form to a world->world form
  Matrix swapy1(4,4), samp1(4,4);
  Matrix swapy2(4,4), samp2(4,4);
  Identity(swapy1);
  swapy1(2,2)=-1;
  swapy2=swapy1;
  swapy1(2,4)=initvol.ysize()-1.0;  // corrected version
  swapy2(2,4)=finalimsize(2)-1.0;   // corrected version
  Identity(samp1);
  samp1(1,1) = initvol.xdim();  
  samp1(2,2) = initvol.ydim();
  samp1(3,3) = initvol.zdim();
  samp2 = finalvoxsize;
  samp2(1,4)=0; samp2(2,4)=0; samp2(3,4)=0;  // translation can be ignored
  samp2(1,1)=fabs(samp2(1,1));    // swapping info is not used
  samp2(2,2)=fabs(samp2(2,2)); 
  samp2(3,3)=fabs(samp2(3,3));
  target = samp2 * swapy2 * medxmat * swapy1 * samp1.i(); // correct
  return 0;
}


template <class T>
int world2medx(Matrix &medxmat, const Matrix& worldmat, 
	       const ColumnVector& finalimsize, const Matrix& finalvoxsize,
	       const volume<T>& initvol) 
{
  Tracer tr("world2medx");
  // convert the world->world form to a voxel->voxel MEDx form 
  Matrix swapy1(4,4), samp1(4,4);
  Matrix swapy2(4,4), samp2(4,4);
  Identity(swapy1);
  swapy1(2,2)=-1;
  swapy2=swapy1;
  swapy1(2,4)=initvol.ysize()-1.0;  // corrected version
  swapy2(2,4)=finalimsize(2)-1.0;   // corrected version
  Identity(samp1);
  samp1(1,1) = initvol.xdim();  
  samp1(2,2) = initvol.ydim();
  samp1(3,3) = initvol.zdim();
  samp2 = finalvoxsize;
  samp2(1,4)=0; samp2(2,4)=0; samp2(3,4)=0;  // translation can be ignored
  samp2(1,1)=fabs(samp2(1,1));    // swapping info is not used
  samp2(2,2)=fabs(samp2(2,2)); 
  samp2(3,3)=fabs(samp2(3,3));
  medxmat = swapy2 * samp2.i() * worldmat * samp1 * swapy1;
  return 0;
}


template <class T>
int minc2world(Matrix &target, const Matrix& mincmat, 
	       const volume<T>& initvol, const volume<T>& finalvol)
{
  Tracer tr("minc2world");
  // convert the voxel->voxel MINC form to a world->world form
  Matrix trans1(4,4), trans2(4,4), samp1(4,4), samp2(4,4), 
         flipx1(4,4), flipx2(4,4);

  Identity(trans1);
  trans1.SubMatrix(1,3,4,4) = initvol.getorigin();

  Identity(trans2);
  trans2.SubMatrix(1,3,4,4) = finalvol.getorigin();

  Identity(samp1);
  samp1(1,1) = fabs(initvol.xdim());  
  samp1(2,2) = fabs(initvol.ydim());
  samp1(3,3) = fabs(initvol.zdim());

  Identity(samp2);
  samp2(1,1) = fabs(finalvol.xdim());  
  samp2(2,2) = fabs(finalvol.ydim());
  samp2(3,3) = fabs(finalvol.zdim());
  
  Identity(flipx1);
  flipx1(1,1) = -1.0;
  flipx1(1,4) = initvol.xsize() - 1.0;

  Identity(flipx2);
  flipx2(1,1) = -1.0;
  flipx2(1,4) = finalvol.xsize() - 1.0;

  target = samp2 * flipx2 * trans2 * samp2.i() * mincmat * samp1 *
                  trans1.i() * flipx1 * samp1.i();
  return 0;
}


template <class T>
int read_medx_matrix(Matrix &target, ifstream& matfile, const volume<T>& invol)
{
  Tracer tr("read_medx_matrix");
  string curword;
  Matrix id4(4,4), gmat, imsize(1,3), voxsize(4,4), medxmat(4,4);
  Identity(id4);  // set to be the identity matrix for checking
  medxmat = id4;   // default return value
  imsize = 1.0;
  voxsize = id4;
  while (!matfile.eof()) {
    matfile >> curword;
    if ( (curword == "/outputsize") ) {
      get_medx_small_matrix(imsize,matfile);
    }
    if ( (curword == "/outputusermatrix") ) {
      get_medx_small_matrix(voxsize,matfile);
    }
    if ( (curword == "/GenericReslice")  || (curword == "/ShadowTransform")
	 || (curword == "/MotionCorrectionReslice") 
	 || (curword == "/UserTransformation") 
	 || (curword == "/IntoTalairachSpace") ) {
      if (get_medx_matrix(gmat,matfile)==0) {  // only accept good returns
	if (SumSquare(gmat - id4) > 0.001) // only accept non-identities
	  medxmat = gmat;
      }
    }
  }
  ColumnVector imsizev(3);
  imsizev(1) = imsize(1,1);
  imsizev(2) = imsize(1,2);
  imsizev(3) = imsize(1,3);
  medx2world(target,medxmat,imsizev,voxsize,invol);
  return 0;
}


template <class T>
int read_minc_matrix(Matrix &target, ifstream& matfile, const volume<T>& invol,
		     const volume<T>& finalvol)
{
  Tracer tr("read_minc_matrix");
  string curword;
  Matrix gmat, id4(4,4);
  Identity(id4);
  target = id4;  // default return value
  while (!matfile.eof()) {
    matfile >> curword;
    if ( (curword == "Linear_Transform") ) {
      get_minc_matrix(gmat,matfile);
    }
    curword="";
  }
  minc2world(target,gmat,invol,finalvol);
  return 0;
}


template <class T>
int read_matrix(Matrix &target, const string& filename, const volume<T>& invol,
		const volume<T>& finalvol)
{
  Tracer tr("read_matrix");
  if ( filename.size()<1 ) return -1;
  ifstream matfile(filename.c_str());
  if (!matfile) { 
    cerr << "Could not open matrix file " << filename << endl;
    return -1;
  }
  string firstline;
  matfile >> firstline;
  matfile.seekg(0,ios::beg);
  int rval=0;
  if (firstline == "%!VEST-Transformations")
    {
      rval = read_medx_matrix(target,matfile,invol);
    } else if (firstline == "MNI") {
      if (finalvol.xsize()>0) {
	rval = read_minc_matrix(target,matfile,invol,finalvol);
      } else {
	cerr << "Cannot read MINC transform files with finalvol" << endl;
	rval = -1;
      }
    } else {
      rval = read_ascii_matrix(target,matfile);
    }
  matfile.close();
  return rval;
}

template <class T>
int read_matrix(Matrix &target, const string& filename, const volume<T>& invol)
{
  volume<T> dummy(0,0,0);
  return read_matrix(target,filename,invol,dummy);
}


template <class T>
int write_medx_matrix(const Matrix& worldmat, const string& filename, 
		      const volume<T>& initvol, 
		      const volume<T>& finalvol, 
		      const string& mtype, const string& reffname)
{
  Tracer tr("write_medx_matrix");
  if ( (filename.size()<1) ) return -1;
  ofstream matfile(filename.c_str());
  if (!matfile) { 
    cerr << "Could not open file " << filename << " for writing" << endl;
    return -1;
  }

  // convert from world->world transform to the voxel->voxel MEDx transform
  Matrix swapy1(4,4), swapy2(4,4), samp1(4,4), samp2(4,4), medxtrans(4,4);
  Identity(samp1);  Identity(samp2);  Identity(swapy1);  Identity(swapy2);
  swapy1(2,2) = -1.0;
  swapy2(2,2) = -1.0;
  swapy1(2,4) = initvol.ysize()-1.0;  // corrected version
  swapy2(2,4) = finalvol.ysize()-1.0; // corrected version
  samp1(1,1) = initvol.xdim();  
  samp1(2,2) = initvol.ydim();
  samp1(3,3) = initvol.zdim();
  samp2(1,1) = finalvol.xdim();  
  samp2(2,2) = finalvol.ydim();
  samp2(3,3) = finalvol.zdim();
  medxtrans = swapy2 * samp2.i() * worldmat * samp1 * swapy1;


  string mattype;
  if ( (mtype.size()<1) || // default
       (mtype[0] == 'g') || (mtype[0] == 'G') )  
    { mattype="GenericReslice"; }
  if ( (mtype[0] == 's') || (mtype[0] == 'S') )  
    { mattype="ShadowTransform"; }
  if ( (mtype[0] == 't') || (mtype[0] == 'T') )  
    { mattype="IntoTalairachSpace"; }
  if ( (mtype[0] == 'u') || (mtype[0] == 'U') )  
    { mattype="UserTransformation"; }
  if ( (mtype[0] == 'm') || (mtype[0] == 'M') || 
       (mtype[0] == 'a') || (mtype[0] == 'A') )
    { mattype="MotionCorrectionReslice"; }

  matfile << "%!VEST-Transformations" << endl;
  matfile << "<<" << endl;
  matfile << "    /" << mattype << "     <<" << endl;
  put_medx_matrix(matfile,"matrix",medxtrans);
  matfile << "        /order  1" << endl;

  Matrix imsize(1,3);
  imsize(1,1) = finalvol.xsize();
  imsize(1,2) = finalvol.ysize();
  imsize(1,3) = finalvol.zsize();
  put_medx_matrix(matfile,"outputsize",imsize);

  Matrix oumat(4,4);
  get_outputusermat(reffname,oumat);
  put_medx_matrix(matfile,"outputusermatrix",oumat);

  if (mattype == "MotionCorrectionReslice") {
    matfile << "        /outputunits    (mm)" << endl;
    matfile << "        /talairachcalibrated?   false" << endl;
  }
  if (mattype == "IntoTalairachSpace") {
    matfile << "        /talairachcalibrated?   true" << endl;
  }
  matfile << "    >>" << endl;
  matfile << ">>" << endl;
  matfile.close();
  return 0;
}


}

#endif

